# Copyright 2025 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import http.client as http
import os
import time

from oslo_config import cfg
from oslo_serialization import jsonutils

from glance.tests import functional
from glance.tests.functional import ft_utils as func_utils
from glance.tests.functional.v2 import test_images


CONF = cfg.CONF

TENANT1 = test_images.TENANT1
TENANT2 = test_images.TENANT2
TENANT3 = test_images.TENANT3
TENANT4 = test_images.TENANT4


class TestImagesMultipleBackend(functional.SynchronousAPIBase):

    def setUp(self):
        super(TestImagesMultipleBackend, self).setUp()
        self.api_methods = test_images.ImageAPIHelper(
            self.api_get, self.api_post, self.api_put, self.api_delete)

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': TENANT1,
            'X-Roles': 'reader,member',
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def _verify_available_stores(self, expected_stores):
        """Verify that the available stores do not contain the staging
        store and have valid IDs.

        This method checks the list of expected stores against the available
        stores and asserts that the staging store is not present.
        """
        path = '/v2/info/stores'
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)

        discovery_stores = jsonutils.loads(response.text)['stores']

        for store in discovery_stores:
            self.assertIn('id', store)
            self.assertIn(store['id'], expected_stores)
            self.assertFalse(store['id'].startswith("os_glance_"))

    def test_image_import_using_glance_direct(self):
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # glance-direct should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method()

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Upload some image data to staging area
        image_data = b'ZZZZZ'
        self.api_methods.stage_image_data(image_id, data=image_data)

        # Verify image is in uploading state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  size=len(image_data),
                                                  status='uploading')

        # Import image to store
        data = {
            'method': {
                'name': 'glance-direct'
            }
        }
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id, data=image_data)

        # Ensure the size is updated to reflect the data uploaded
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1'])

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_image_import_using_glance_direct_different_backend(self):
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # glance-direct should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method()

        # store1 and store2 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(expected_stores=available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Upload some image data to staging area
        image_data = b'ZZZZZ'
        self.api_methods.stage_image_data(image_id, data=image_data)

        # Verify image is in uploading state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  size=len(image_data),
                                                  status='uploading')

        # Import image to store2 store (other than default backend)
        headers = self._headers({
            'content-type': 'application/json',
            'X-Roles': 'admin',
            'X-Image-Meta-Store': 'store2'
        })
        data = {'method': {
            'name': 'glance-direct'
        }}
        self.api_methods.import_image(image_id, data=data,
                                      headers=headers)
        self.api_methods.verify_image_import_status(image_id, data=image_data)

        # Ensure the size is updated to reflect the data uploaded and
        # image is imported to store2
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store2'])

        # Deleting image should work
        self.api_methods.delete_image(image_id)

        # Image list should now be empty
        self.api_methods.verify_empty_image_list

    def test_image_import_using_web_download(self):
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # web-download should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='web-download')

        # store1 and store2 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(expected_stores=available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {'method': {
            'name': 'web-download',
            'uri': image_data_uri
        }}
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # Ensure the size is updated to reflect the data imported and
        # image is imported to default backend
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1'])

        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_image_import_using_web_download_different_backend(self):
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # web-download should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='web-download')

        # store1 and store2 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(expected_stores=available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {'method': {
            'name': 'web-download',
            'uri': image_data_uri
        }}
        # Import image to store
        # Import image to store
        headers = self._headers({
            'content-type': 'application/json',
            'X-Roles': 'admin',
            'X-Image-Meta-Store': 'store2'
        })
        self.api_methods.import_image(image_id, data=data, headers=headers)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # Ensure the size is updated to reflect the data imported and
        # image is imported to default backend
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store2'])

        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_image_import_multi_stores(self):
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        # Image list should be empty
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # web-download should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='web-download')

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {
            'method': {
                'name': 'web-download',
                'uri': image_data_uri
            },
            'stores': ['store1', 'store2']
        }
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # kill the local http server
        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # Ensure image is created in the two stores
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1', 'store2'])

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_copy_image_lifecycle(self):
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # copy-image should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='copy-image')

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {'method': {
            'name': 'web-download',
            'uri': image_data_uri
        }}
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # Ensure the size is updated to reflect the data imported and
        # image is imported to default backend
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1'])

        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # Ensure image has one task associated with it
        path = '/v2/images/%s/tasks' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        tasks = jsonutils.loads(response.text)['tasks']
        self.assertEqual(1, len(tasks))
        for task in tasks:
            self.assertEqual(image_id, task['image_id'])
            user_id = response.request.headers.get(
                'X-User-Id')
            self.assertEqual(user_id, task['user_id'])
            self.assertEqual(self.api_methods.import_reqid, task['request_id'])

        # Copy newly created image to store2 and store3 stores
        path = '/v2/images/%s/import' % image_id
        headers = self._headers({
            'content-type': 'application/json',
            'X-Roles': 'admin'
        })

        data = {
            'method': {'name': 'copy-image'},
            'stores': ['store2', 'store3']
        }
        response = self.api_post(path, headers=headers, json=data)
        self.assertEqual(http.ACCEPTED, response.status_code)
        copy_reqid = response.headers['X-Openstack-Request-Id']

        # Verify image is copied
        # NOTE(abhishekk): As import is a async call we need to provide
        # some timelap to complete the call.
        path = '/v2/images/%s' % image_id
        func_utils.wait_for_copying(request_path=path,
                                    request_headers=self._headers(),
                                    stores=['store2', 'store3'],
                                    max_sec=40,
                                    delay_sec=0.2,
                                    start_delay_sec=1,
                                    api_get_method=self.api_get)

        # Ensure image is copied to the store2 and store3 store and available
        # in all 3 stores now
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1', 'store2', 'store3'])

        # Ensure image has two tasks associated with it
        path = '/v2/images/%s/tasks' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        tasks = jsonutils.loads(response.text)['tasks']
        self.assertEqual(2, len(tasks))
        expected_reqids = [copy_reqid, self.api_methods.import_reqid]
        for task in tasks:
            self.assertEqual(image_id, task['image_id'])
            user_id = response.request.headers.get(
                'X-User-Id')
            self.assertEqual(user_id, task['user_id'])
            self.assertEqual(expected_reqids.pop(), task['request_id'])

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_copy_image_revert_lifecycle(self):
        # Test if copying task fails in between then the rollback
        # should delete the data from only stores to which it is
        # copied and not from the existing stores.
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()
        # Image list should be empty
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # copy-image should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='copy-image')

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {'method': {
            'name': 'web-download',
            'uri': image_data_uri
        }}
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # Ensure the size is updated to reflect the data imported and
        # image is imported to default backend
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1'])

        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # NOTE(abhishekk): Deleting store3 image directory to trigger the
        # failure, so that we can verify that revert call does not delete
        # the data from existing stores
        # NOTE(danms): Do this before we start the import, on a later store,
        # which will cause that store to fail after we have already completed
        # the first one.
        os.rmdir(self.test_dir + "/store3")

        # Copy newly created image to store2 and store3 store
        path = '/v2/images/%s/import' % image_id
        headers = self._headers({
            'content-type': 'application/json',
            'X-Roles': 'admin'
        })
        data = {
            'method': {'name': 'copy-image'},
            'stores': ['store2', 'store3']
        }
        response = self.api_post(path, headers=headers, json=data)
        self.assertEqual(http.ACCEPTED, response.status_code)

        def poll_callback(image):
            # NOTE(danms): We need to wait for the specific
            # arrangement we're expecting, which is that file3 has
            # failed, nothing else is importing, and store2 has been
            # removed from stores by the revert.
            return not (image['os_glance_importing_to_stores'] == '' and
                        image['os_glance_failed_import'] == 'store3' and
                        image['stores'] == 'store1')

        func_utils.poll_entity('/v2/images/%s' % image_id,
                               self._headers(),
                               poll_callback,
                               api_get_method=self.api_get)

        # Here we check that the failure of 'store2' caused 'store2' to
        # be removed from image['stores'], and that 'store3' is reported
        # as failed in the appropriate status list. Since the import
        # started with 'store1' being populated, that should remain,
        # but 'store2' should be reverted/removed.
        response = self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1'],
            rejected_stores=['store2', 'store3'])
        fail_key = 'os_glance_failed_import'
        pend_key = 'os_glance_importing_to_stores'
        self.assertEqual('store3', response[fail_key])
        self.assertEqual('', response[pend_key])

        # Copy newly created image to store2 and store3 stores and
        # all_stores_must_succeed set to false.
        headers = self._headers({
            'content-type': 'application/json',
            'X-Roles': 'admin'
        })

        data = {
            'method': {'name': 'copy-image'},
            'stores': ['store2', 'store3'],
            'all_stores_must_succeed': False
        }

        for i in range(0, 5):
            response = self.api_post(path, headers=headers, json=data)
            if response.status_code != http.CONFLICT:
                break
            # We might race with the revert of the previous task and do not
            # really have a good way to make sure that it's done. In order
            # to make sure we tolerate the 409 possibility when import
            # locking is added, gracefully wait a few times before failing.
            time.sleep(1)

        self.assertEqual(http.ACCEPTED, response.status_code)

        # Verify image is copied
        # NOTE(abhishekk): As import is a async call we need to provide
        # some timelap to complete the call.
        path = '/v2/images/%s' % image_id
        func_utils.wait_for_copying(request_path=path,
                                    request_headers=self._headers(),
                                    stores=['store2'],
                                    max_sec=10,
                                    delay_sec=0.2,
                                    start_delay_sec=1,
                                    failure_scenario=True,
                                    api_get_method=self.api_get)

        # Ensure data is not deleted from existing stores as well as
        # from the stores where it is copied successfully
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1', 'store2'],
            rejected_stores=['store3'])

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_image_import_multi_stores_specifying_all_stores(self):
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()
        # Image list should be empty
        self.start_server()

        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # copy-image should be available in discovery response
        self.api_methods.verify_discovery_includes_import_method(
            method='copy-image')

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Verify image is in queued state and checksum is None
        func_utils.verify_image_hashes_and_status(self, image_id,
                                                  status='queued')

        # Start http server locally
        image_data_uri = self.api_methods.start_http_server_and_get_uri()
        data = {
            'method': {
                'name': 'web-download',
                'uri': image_data_uri,
                'all_stores': True
            },
            'all_stores': True
        }
        self.api_methods.import_image(image_id, data=data)
        self.api_methods.verify_image_import_status(image_id,
                                                    data=image_data_uri)

        # Ensure image is imported to all available stores
        self.api_methods.verify_image_stores(
            image_id, expected_stores=['store1', 'store2', 'store3'])

        # kill the local http server
        self.api_methods.httpd.shutdown()
        self.api_methods.httpd.server_close()

        # Clean up: Delete the image and verify the list is empty
        self.api_methods.delete_image(image_id)
        self.api_methods.verify_empty_image_list()

    def test_image_lifecycle(self):
        self.start_server()
        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image (with two deployer-defined properties)
        additional_properties = {
            'protected': True,
            'foo': 'bar',
            'abc': 'xyz'
        }
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores,
            additional_properties=additional_properties)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Try to download data before its uploaded
        path = '/v2/images/%s/file' % image_id
        headers = self._headers()
        response = self.api_get(path, headers=headers)
        self.assertEqual(http.NO_CONTENT, response.status_code)

        # Upload some image data
        image_data = b'OpenStack Rules, Other Clouds Drool'
        self.api_methods.upload_and_verify(image_id, image_data)

        # Ensure image is created in default backend
        self.api_methods.verify_image_stores(image_id,
                                             expected_stores=['store1'])

        # Try to download the data that was just uploaded
        expect_c = hashlib.md5(image_data, usedforsecurity=False).hexdigest()
        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        self.assertEqual(expect_c, response.headers['Content-MD5'])
        self.assertEqual(image_data.decode('utf-8'), response.text)

        # Ensure the size is updated to reflect the data uploaded
        self.api_methods.verify_image_size(
            image_id, expected_size=len(image_data))

        # Unprotect image for deletion
        path = '/v2/images/%s' % image_id
        media_type = 'application/openstack-images-v2.1-json-patch'
        headers = self._headers({'content-type': media_type})
        doc = [{'op': 'replace', 'path': '/protected', 'value': False}]
        response = self.api_patch(path, doc, headers=headers)
        self.assertEqual(http.OK, response.status_code, response.text)

        # Deletion should work. Deleting image
        self.api_methods.delete_image(image_id)

        # This image should be no longer be directly accessible
        self.api_methods.assert_image_not_found(image_id)

        # And neither should its data
        path = '/v2/images/%s/file' % image_id
        headers = self._headers()
        response = self.api_get(path, headers=headers)
        self.assertEqual(http.NOT_FOUND, response.status_code)

        # Image list should now be empty
        self.api_methods.verify_empty_image_list

    def test_image_lifecycle_different_backend(self):
        self.start_server()
        # Image list should be empty
        self.api_methods.verify_empty_image_list()

        # store1, store2 and store3 should be available in discovery response
        available_stores = ['store1', 'store2', 'store3']
        self._verify_available_stores(available_stores)

        # Create an image (with two deployer-defined properties)
        additional_properties = {
            'protected': True,
            'foo': 'bar',
            'abc': 'xyz'
        }
        image_id = self.api_methods.create_and_verify_image(
            name='image-1', type='kernel',
            disk_format='aki', container_format='aki',
            verify_stores=True, available_stores=available_stores,
            additional_properties=additional_properties)

        # Image list should now have one entry
        self.api_methods.verify_image_list_contains(image_id,
                                                    expected_count=1)

        # Try to download data before its uploaded
        path = '/v2/images/%s/file' % image_id
        headers = self._headers()
        response = self.api_get(path, headers=headers)
        self.assertEqual(http.NO_CONTENT, response.status_code)

        # Upload some image data
        image_data = b'just a passing glance'
        headers = self._headers({
            'Content-Type': 'application/octet-stream',
            'X-Image-Meta-Store': 'store2'
        })
        self.api_methods.upload_and_verify(image_id, image_data,
                                           headers=headers)

        # Ensure image is created in default backend
        self.api_methods.verify_image_stores(image_id,
                                             expected_stores=['store2'])

        # Try to download the data that was just uploaded
        expect_c = hashlib.md5(image_data, usedforsecurity=False).hexdigest()
        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        self.assertEqual(expect_c, response.headers['Content-MD5'])
        self.assertEqual(image_data.decode('utf-8'), response.text)

        # Ensure the size is updated to reflect the data uploaded
        self.api_methods.verify_image_size(
            image_id, expected_size=len(image_data))

        # Unprotect image for deletion
        path = '/v2/images/%s' % image_id
        media_type = 'application/openstack-images-v2.1-json-patch'
        headers = self._headers({'content-type': media_type})
        doc = [{'op': 'replace', 'path': '/protected', 'value': False}]
        response = self.api_patch(path, doc, headers=headers)
        self.assertEqual(http.OK, response.status_code, response.text)

        # Deletion should work. Deleting image
        self.api_methods.delete_image(image_id)

        # This image should be no longer be directly accessible
        self.api_methods.assert_image_not_found(image_id)

        # And neither should its data
        path = '/v2/images/%s/file' % image_id
        headers = self._headers()
        response = self.api_get(path, headers=headers)
        self.assertEqual(http.NOT_FOUND, response.status_code)

        # Image list should now be empty
        self.api_methods.verify_empty_image_list



# Copyright 2019 RedHat Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import utils as db_utils

from glance.db.sqlalchemy.alembic_migrations import data_migrations
from glance.tests.functional.db import test_migrations
import glance.tests.utils as test_utils


class TestTrainMigrate01Mixin(test_migrations.AlembicMigrationsMixin):

    def _get_revisions(self, config):
        return test_migrations.AlembicMigrationsMixin._get_revisions(
            self, config, head='train_expand01')

    def _pre_upgrade_train_expand01(self, engine):
        images = db_utils.get_table(engine, 'images')
        image_locations = db_utils.get_table(engine, 'image_locations')
        now = datetime.datetime.now()

        # inserting a public image record
        image_1 = dict(deleted=False,
                       created_at=now,
                       status='active',
                       min_disk=0,
                       min_ram=0,
                       visibility='public',
                       id='image_1')
        with engine.connect() as conn, conn.begin():
            conn.execute(images.insert().values(image_1))

        image_2 = dict(deleted=False,
                       created_at=now,
                       status='active',
                       min_disk=0,
                       min_ram=0,
                       visibility='public',
                       id='image_2')
        with engine.connect() as conn, conn.begin():
            conn.execute(images.insert().values(image_2))

        # adding records to image_locations tables
        temp = dict(deleted=False,
                    created_at=now,
                    image_id='image_1',
                    value='image_location_1',
                    meta_data='{"backend": "fast"}',
                    id=1)
        with engine.connect() as conn, conn.begin():
            conn.execute(image_locations.insert().values(temp))

        temp = dict(deleted=False,
                    created_at=now,
                    image_id='image_2',
                    value='image_location_2',
                    meta_data='{"backend": "cheap"}',
                    id=2)
        with engine.connect() as conn, conn.begin():
            conn.execute(image_locations.insert().values(temp))

    def _check_train_expand01(self, engine, data):
        image_locations = db_utils.get_table(engine, 'image_locations')

        # check that meta_data has 'backend' key for existing image_locations
        with engine.connect() as conn:
            rows = conn.execute(
                image_locations.select().order_by(image_locations.c.id)
            ).fetchall()

        self.assertEqual(2, len(rows))
        for row in rows:
            self.assertIn('"backend":', row.meta_data)

        # run data migrations
        data_migrations.migrate(engine, release='train')

        # check that meta_data has 'backend' key replaced with 'store'
        with engine.connect() as conn:
            rows = conn.execute(
                image_locations.select().order_by(image_locations.c.id)
            ).fetchall()

        self.assertEqual(2, len(rows))
        for row in rows:
            self.assertNotIn('"backend":', row.meta_data)
            self.assertIn('"store":', row.meta_data)


class TestTrainMigrate01MySQL(
    TestTrainMigrate01Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestTrain01PostgresSQL(
    TestTrainMigrate01Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class TestTrainMigrate01_EmptyDBMixin(test_migrations.AlembicMigrationsMixin):
    """This mixin is used to create an initial glance database and upgrade it
    up to the train_expand01 revision.
    """
    def _get_revisions(self, config):
        return test_migrations.AlembicMigrationsMixin._get_revisions(
            self, config, head='train_expand01')

    def _pre_upgrade_train_expand01(self, engine):
        # New/empty database
        pass

    def _check_train_expand01(self, engine, data):
        images = db_utils.get_table(engine, 'images')

        # check that there are no rows in the images table
        with engine.connect() as conn:
            rows = conn.execute(
                images.select().order_by(images.c.id)
            ).fetchall()

        self.assertEqual(0, len(rows))

        # run data migrations
        data_migrations.migrate(engine)


class TestTrainMigrate01_EmptyDBMySQL(
    TestTrainMigrate01_EmptyDBMixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestTrainMigrate01_PySQL(
    TestTrainMigrate01_EmptyDBMixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



# Copyright 2016 Rackspace
# Copyright 2016 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import os.path
import pkgutil

from glance.common import exception
from glance.db import migration as db_migrations
from glance.db.sqlalchemy import api as db_api


def _find_migration_modules(release):
    migrations = list()
    for _, module_name, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
        if module_name.startswith(release):
            migrations.append(module_name)

    migration_modules = list()
    for migration in sorted(migrations):
        module = importlib.import_module('.'.join([__package__, migration]))
        has_migrations_function = getattr(module, 'has_migrations', None)
        migrate_function = getattr(module, 'migrate', None)

        if has_migrations_function is None or migrate_function is None:
            raise exception.InvalidDataMigrationScript(script=module.__name__)

        migration_modules.append(module)

    return migration_modules


def _run_migrations(engine, migrations):
    rows_migrated = 0
    for migration in migrations:
        if migration.has_migrations(engine):
            rows_migrated += migration.migrate(engine)

    return rows_migrated


def has_pending_migrations(engine=None, release=db_migrations.CURRENT_RELEASE):
    if not engine:
        engine = db_api.get_engine()

    migrations = _find_migration_modules(release)
    if not migrations:
        return False
    return any([x.has_migrations(engine) for x in migrations])


def migrate(engine=None, release=db_migrations.CURRENT_RELEASE):
    if not engine:
        engine = db_api.get_engine()

    migrations = _find_migration_modules(release)
    rows_migrated = _run_migrations(engine, migrations)
    return rows_migrated



# Copyright 2022 OVHcloud
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

from glance.async_ import utils
import glance.common.exception
from glance.tests.unit import base


class TestGetGlanceEndpoint(base.IsolatedUnitTest):

    def setUp(self):
        super(TestGetGlanceEndpoint, self).setUp()

        self.service_catalog = [
            {
                'endpoints': [
                    {
                        'adminURL': 'http://localhost:8080/',
                        'region': 'RegionOne',
                        'internalURL': 'http://internalURL/',
                        'publicURL': 'http://publicURL/',
                    },
                ],
                'type': 'object-store',
            },
            {
                'endpoints': [
                    {
                        'adminURL': 'http://localhost:8080/',
                        'region': 'RegionOne',
                        'internalURL': 'http://RegionOneInternal/',
                        'publicURL': 'http://RegionOnePublic/',
                    },
                ],
                'type': 'image',
            },
            {
                'endpoints': [
                    {
                        'adminURL': 'http://localhost:8080/',
                        'region': 'RegionTwo',
                        'internalURL': 'http://RegionTwoInternal/',
                        'publicURL': 'http://RegionTwoPublic/',
                    },
                ],
                'type': 'image',
            }
        ]

        self.context = mock.MagicMock(service_catalog=self.service_catalog)

    def test_return_matching_glance_endpoint(self):
        self.assertEqual(utils.get_glance_endpoint(self.context,
                                                   'RegionOne',
                                                   'public'),
                         'http://RegionOnePublic/')
        self.assertEqual(utils.get_glance_endpoint(self.context,
                                                   'RegionTwo',
                                                   'internal'),
                         'http://RegionTwoInternal/')

    def test_glance_endpoint_not_found(self):
        self.assertRaises(glance.common.exception.GlanceEndpointNotFound,
                          utils.get_glance_endpoint, self.context,
                          'RegionThree', 'public')



# Copyright 2020 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys
from unittest import mock

import fixtures

from glance.cmd import cache_manage
from glance.image_cache import client as cache_client
from glance.tests import utils as test_utils


UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'


class TestCacheManage(test_utils.BaseTestCase):

    def setUp(self):
        super(TestCacheManage, self).setUp()

    def _main_test_helper(self, argv, result=cache_manage.SUCCESS):
        self.useFixture(fixtures.MonkeyPatch('sys.argv', argv))
        with mock.patch.object(cache_client, 'get_client'):
            with mock.patch.object(sys, 'exit') as mock_exit:
                cache_manage.main()
            mock_exit.assert_called_once_with(result)

    def test_list_cached_images(self):
        self._main_test_helper(['glance.cmd.cache_manage', 'list-cached'])

    def test_list_queued_images(self):
        self._main_test_helper(['glance.cmd.cache_manage', 'list-queued'])

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_queue_image(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'queue-image', UUID1])
        self.assertEqual(1, mock_user_confirm.call_count)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_queue_image_invalid_image_id(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage', 'queue-image',
                                'fake_id'],
                               result=cache_manage.FAILURE)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_queued_image(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-queued-image', UUID1])
        self.assertEqual(1, mock_user_confirm.call_count)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_queued_image_invalid_image_id(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-queued-image',
                                'fake_id'],
                               result=cache_manage.FAILURE)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_cached_image(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-cached-image', UUID1])
        self.assertEqual(1, mock_user_confirm.call_count)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_cached_image_invalid_image_id(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-cached-image',
                                'fake_id'],
                               result=cache_manage.FAILURE)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_all_queued_image(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-all-queued-images'])
        self.assertEqual(1, mock_user_confirm.call_count)

    @mock.patch.object(cache_manage, 'user_confirm')
    def test_delete_all_cached_image(self, mock_user_confirm):
        self._main_test_helper(['glance.cmd.cache_manage',
                                'delete-all-cached-images'])
        self.assertEqual(1, mock_user_confirm.call_count)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import oslo_policy.policy
from oslo_utils.fixture import uuidsentinel as uuids

from glance.api import policy
from glance.tests import functional


class TestImageMembersPolicy(functional.SynchronousAPIBase):
    def setUp(self):
        super(TestImageMembersPolicy, self).setUp()
        self.policy = policy.Enforcer(suppress_deprecation_warnings=True)

    def load_data(self, share_image=False):
        output = {}
        path = "/v2/images"
        data = {
            'name': 'shared-image',
            'visibility': 'shared',
        }
        response = self.api_post(path, json=data)
        self.assertEqual(201, response.status_code)
        image_id = response.json['id']
        output['image_id'] = image_id

        if share_image:
            path = '/v2/images/%s/members' % image_id
            data = {
                'member': uuids.random_member
            }
            response = self.api_post(path, json=data)
            member = response.json
            self.assertEqual(200, response.status_code)
            self.assertEqual(image_id, member['image_id'])
            self.assertEqual('pending', member['status'])
            output['member_id'] = member['member_id']

        return output

    def set_policy_rules(self, rules):
        self.policy.set_rules(
            oslo_policy.policy.Rules.from_dict(rules),
            overwrite=True)

    def start_server(self):
        with mock.patch.object(policy, 'Enforcer') as mock_enf:
            mock_enf.return_value = self.policy
            super(TestImageMembersPolicy, self).start_server()

    def test_member_add_basic(self):
        self.start_server()
        output = self.load_data()
        path = '/v2/images/%s/members' % output['image_id']

        # Make sure we can add member to image (can share image)
        data = {
            'member': uuids.random_member
        }
        response = self.api_post(path, json=data)
        self.assertEqual(200, response.status_code)
        member = response.json
        self.assertEqual(output['image_id'], member['image_id'])
        self.assertEqual('pending', member['status'])

        # Now disable add permissions
        self.set_policy_rules({
            'add_member': '!',
            'get_image': '@'
        })

        # Make sure add returns 403
        response = self.api_post(path, json=data)
        self.assertEqual(403, response.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'add_member': '!',
            'get_image': '!'
        })

        # Make sure add returns 404
        response = self.api_post(path, json=data)
        self.assertEqual(404, response.status_code)

    def test_member_update_basic(self):
        self.start_server()
        output = self.load_data(share_image=True)
        path = '/v2/images/%s/members/%s' % (output['image_id'],
                                             output['member_id'])
        # Make sure we can update image membership
        data = {
            'status': 'accepted'
        }
        response = self.api_put(path, json=data)
        self.assertEqual(200, response.status_code)
        member = response.json
        self.assertEqual(output['image_id'], member['image_id'])
        self.assertEqual('accepted', member['status'])

        # Now disable modify permissions
        self.set_policy_rules({
            'modify_member': '!',
            'get_image': '@'
        })

        # Make sure update returns 403
        response = self.api_put(path, json=data)
        self.assertEqual(403, response.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'modify_member': '!',
            'get_image': '!',
            'get_member': '@'
        })
        # image owner is not allowed to update image membership so
        # passing different project in headers
        headers = self._headers({
            'X-Tenant-Id': 'fake-tenant-id',
        })
        response = self.api_put(path, headers=headers, json=data)
        self.assertEqual(404, response.status_code)

    def test_member_list_basic(self):
        self.start_server()
        output = self.load_data(share_image=True)
        path = '/v2/images/%s/members' % (output['image_id'])
        # # Make sure we can list image members
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(response.json['members']))

        # Now disable list permissions
        self.set_policy_rules({
            'get_members': '!',
            'get_image': '@',
        })

        # Make sure get returns 403
        response = self.api_get(path)
        self.assertEqual(403, response.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'get_members': '!',
            'get_image': '!',
        })

        # Make sure get returns 404
        response = self.api_get(path)
        self.assertEqual(404, response.status_code)

        # Now enable get_members and disable get_member
        self.set_policy_rules({
            'get_members': '@',
            'get_member': '!',
            'get_image': '@',
        })

        # Make sure we get empty list as get_member is disabled
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)
        self.assertEqual(0, len(response.json['members']))

    def test_member_get_basic(self):
        self.start_server()
        output = self.load_data(share_image=True)
        path = '/v2/images/%s/members/%s' % (
            output['image_id'],
            output['member_id'])
        # Make sure we can get member
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)
        member = response.json
        self.assertEqual(output['image_id'], member['image_id'])
        self.assertEqual('pending', member['status'])

        # Now disable get permissions
        self.set_policy_rules({'get_member': '!'})

        # Make sure get returns 404 as we are not exposing it
        response = self.api_get(path)
        self.assertEqual(404, response.status_code)

    def test_member_delete_basic(self):
        self.start_server()
        output = self.load_data(share_image=True)
        path = '/v2/images/%s/members/%s' % (output['image_id'],
                                             output['member_id'])
        # Make sure we can delete image member
        response = self.api_delete(path)
        self.assertEqual(204, response.status_code)
        # Verifiy it is deleted
        response = self.api_get(path)
        self.assertEqual(404, response.status_code)

        # Now disable delete permissions and create image
        # membership again
        self.set_policy_rules({
            'delete_member': '!',
            'add_member': '@',
            'get_image': '@'
        })
        add_path = '/v2/images/%s/members' % output['image_id']
        data = {
            'member': uuids.random_member
        }
        response = self.api_post(add_path, json=data)
        self.assertEqual(200, response.status_code)

        # Make sure delete returns 403
        response = self.api_delete(path)
        self.assertEqual(403, response.status_code)

        # Now disable delete_member and get_image permission and make
        # sure you will get 404 Not Found
        self.set_policy_rules({
            'delete_member': '!',
            'get_image': '!',
            'get_member': '@'
        })
        response = self.api_delete(path)
        self.assertEqual(404, response.status_code)

    def test_image_sharing_not_allowed(self):
        # This test verifies that image having visibility other than
        # shared is not allowed for sharing
        self.start_server()
        path = "/v2/images"
        for visibility in ('community', 'private', 'public'):
            data = {
                'name': '%s-image' % visibility,
                'visibility': visibility,
            }
            # create image
            response = self.api_post(path, json=data)
            image = response.json
            self.assertEqual(201, response.status_code)
            self.assertEqual(visibility, image['visibility'])

            # Sharing image should return 403 response
            member_path = '/v2/images/%s/members' % image['id']
            data = {
                'member': uuids.random_member
            }
            response = self.api_post(member_path, json=data)
            self.assertEqual(403, response.status_code)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.client as http
import urllib.parse as urlparse

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import webob.exc
from wsme.rest import json

from glance.api import policy
from glance.api.v2.model.metadef_namespace import Namespace
from glance.api.v2.model.metadef_namespace import Namespaces
from glance.api.v2.model.metadef_object import MetadefObject
from glance.api.v2.model.metadef_property_type import PropertyType
from glance.api.v2.model.metadef_resource_type import ResourceTypeAssociation
from glance.api.v2.model.metadef_tag import MetadefTag
from glance.api.v2 import policy as api_policy
from glance.common import exception
from glance.common import utils
from glance.common import wsgi
from glance.common import wsme_utils
import glance.db
import glance.gateway
from glance.i18n import _, _LE
import glance.notifier
import glance.schema

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class NamespaceController(object):
    def __init__(self, db_api=None, policy_enforcer=None, notifier=None):
        self.db_api = db_api or glance.db.get_api()
        self.policy = policy_enforcer or policy.Enforcer()
        self.notifier = notifier or glance.notifier.Notifier()
        self.gateway = glance.gateway.Gateway(db_api=self.db_api,
                                              notifier=self.notifier,
                                              policy_enforcer=self.policy)
        self.ns_schema_link = '/v2/schemas/metadefs/namespace'
        self.obj_schema_link = '/v2/schemas/metadefs/object'
        self.tag_schema_link = '/v2/schemas/metadefs/tag'

    def index(self, req, marker=None, limit=None, sort_key='created_at',
              sort_dir='desc', filters=None):
        try:
            ns_repo = self.gateway.get_metadef_namespace_repo(req.context)

            policy_check = api_policy.MetadefAPIPolicy(
                req.context,
                enforcer=self.policy)
            # NOTE(abhishekk): This is just a "do you have permission to
            # list namespace" check. Each namespace is checked against
            # get_metadef_namespace below.
            policy_check.get_metadef_namespaces()

            # NOTE(abhishekk): We also need to fetch resource_types associated
            # with namespaces, so better to check we have permission for the
            # same in advance.
            policy_check.list_metadef_resource_types()

            # Get namespace id
            if marker:
                namespace_obj = ns_repo.get(marker)
                marker = namespace_obj.namespace_id

            database_ns_list = ns_repo.list(
                marker=marker, limit=limit, sort_key=sort_key,
                sort_dir=sort_dir, filters=filters)

            ns_list = [
                ns for ns in database_ns_list if api_policy.MetadefAPIPolicy(
                    req.context, md_resource=ns, enforcer=self.policy).check(
                    'get_metadef_namespace')]

            rs_repo = (
                self.gateway.get_metadef_resource_type_repo(req.context))
            for db_namespace in ns_list:
                # Get resource type associations
                filters = dict()
                filters['namespace'] = db_namespace.namespace
                try:
                    repo_rs_type_list = rs_repo.list(filters=filters)
                except exception.NotFound:
                    # NOTE(danms): If we fail to list resource_types
                    # for this namespace, do not fail the entire
                    # namespace list operation with NotFound.
                    repo_rs_type_list = []
                resource_type_list = [
                    ResourceTypeAssociation.to_wsme_model(
                        resource_type
                    ) for resource_type in repo_rs_type_list]

                if resource_type_list:
                    db_namespace.resource_type_associations = (
                        resource_type_list)

            namespace_list = [Namespace.to_wsme_model(
                db_namespace,
                get_namespace_href(db_namespace),
                self.ns_schema_link) for db_namespace in ns_list]
            namespaces = Namespaces()
            namespaces.namespaces = namespace_list
            if len(namespace_list) != 0 and len(namespace_list) == limit:
                namespaces.next = ns_list[-1].namespace

        except exception.Forbidden as e:
            LOG.debug("User not permitted to retrieve metadata namespaces "
                      "index")
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        return namespaces

    @utils.mutating
    def create(self, req, namespace):
        try:
            namespace_created = False
            # Create Namespace
            ns_factory = self.gateway.get_metadef_namespace_factory(
                req.context)
            ns_repo = self.gateway.get_metadef_namespace_repo(req.context)

            # NOTE(abhishekk): Here we are going to check if user is authorized
            # to create namespace, resource_types, objects, properties etc.
            policy_check = api_policy.MetadefAPIPolicy(
                req.context, enforcer=self.policy)
            policy_check.add_metadef_namespace()

            if namespace.resource_type_associations:
                policy_check.add_metadef_resource_type_association()
            if namespace.objects:
                policy_check.add_metadef_object()
            if namespace.properties:
                policy_check.add_metadef_property()
            if namespace.tags:
                policy_check.add_metadef_tag()

            # NOTE(abhishekk): As we are getting rid of auth layer, this
            # is the place where we should add owner if it is not specified
            # in request.
            kwargs = namespace.to_dict()
            if 'owner' not in kwargs:
                kwargs.update({'owner': req.context.owner})

            new_namespace = ns_factory.new_namespace(**kwargs)
            ns_repo.add(new_namespace)
            namespace_created = True

            # Create Resource Types
            if namespace.resource_type_associations:
                rs_factory = (self.gateway.get_metadef_resource_type_factory(
                    req.context))
                rs_repo = self.gateway.get_metadef_resource_type_repo(
                    req.context)
                for resource_type in namespace.resource_type_associations:
                    new_resource = rs_factory.new_resource_type(
                        namespace=namespace.namespace,
                        **resource_type.to_dict())
                    rs_repo.add(new_resource)

            # Create Objects
            if namespace.objects:
                object_factory = self.gateway.get_metadef_object_factory(
                    req.context)
                object_repo = self.gateway.get_metadef_object_repo(
                    req.context)
                for metadata_object in namespace.objects:
                    new_meta_object = object_factory.new_object(
                        namespace=namespace.namespace,
                        **metadata_object.to_dict())
                    object_repo.add(new_meta_object)

            # Create Tags
            if namespace.tags:
                tag_factory = self.gateway.get_metadef_tag_factory(
                    req.context)
                tag_repo = self.gateway.get_metadef_tag_repo(
                    req.context)
                for metadata_tag in namespace.tags:
                    new_meta_tag = tag_factory.new_tag(
                        namespace=namespace.namespace,
                        **metadata_tag.to_dict())
                    tag_repo.add(new_meta_tag)

            # Create Namespace Properties
            if namespace.properties:
                prop_factory = (self.gateway.get_metadef_property_factory(
                    req.context))
                prop_repo = self.gateway.get_metadef_property_repo(
                    req.context)
                for (name, value) in namespace.properties.items():
                    new_property_type = (
                        prop_factory.new_namespace_property(
                            namespace=namespace.namespace,
                            **self._to_property_dict(name, value)
                        ))
                    prop_repo.add(new_property_type)
        except exception.Invalid as e:
            msg = (_("Couldn't create metadata namespace: %s") % e)
            raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.Forbidden as e:
            self._cleanup_namespace(ns_repo, namespace, namespace_created)
            LOG.debug("User not permitted to create metadata namespace")
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            self._cleanup_namespace(ns_repo, namespace, namespace_created)
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Duplicate as e:
            self._cleanup_namespace(ns_repo, namespace, namespace_created)
            raise webob.exc.HTTPConflict(explanation=e.msg)

        # Return the user namespace as we don't expose the id to user
        new_namespace.properties = namespace.properties
        new_namespace.objects = namespace.objects
        new_namespace.resource_type_associations = (
            namespace.resource_type_associations)
        new_namespace.tags = namespace.tags
        return Namespace.to_wsme_model(new_namespace,
                                       get_namespace_href(new_namespace),
                                       self.ns_schema_link)

    def _to_property_dict(self, name, value):
        # Convert the model PropertyTypes dict to a JSON string
        db_property_type_dict = dict()
        db_property_type_dict['schema'] = json.tojson(PropertyType, value)
        db_property_type_dict['name'] = name
        return db_property_type_dict

    def _cleanup_namespace(self, namespace_repo, namespace, namespace_created):
        if namespace_created:
            try:
                namespace_obj = namespace_repo.get(namespace.namespace)
                namespace_obj.delete()
                namespace_repo.remove(namespace_obj)
                LOG.debug("Cleaned up namespace %(namespace)s ",
                          {'namespace': namespace.namespace})
            except Exception as e:
                msg = (_LE("Failed to delete namespace %(namespace)s."
                           "Exception: %(exception)s"),
                       {'namespace': namespace.namespace,
                        'exception': str(e)})
                LOG.error(msg)

    def show(self, req, namespace, filters=None):
        try:
            # Get namespace
            ns_repo = self.gateway.get_metadef_namespace_repo(
                req.context)
            try:
                namespace_obj = ns_repo.get(namespace)
                policy_check = api_policy.MetadefAPIPolicy(
                    req.context,
                    md_resource=namespace_obj,
                    enforcer=self.policy)
                policy_check.get_metadef_namespace()
            except (exception.Forbidden, webob.exc.HTTPForbidden):
                LOG.debug("User not permitted to show namespace '%s'",
                          namespace)
                # NOTE (abhishekk): Returning 404 Not Found as the
                # namespace is outside of this user's project
                raise webob.exc.HTTPNotFound()

            # NOTE(abhishekk): We also need to fetch resource_types, objects,
            # properties, tags associated with namespace, so better to check
            # whether user has permissions for the same.
            policy_check.list_metadef_resource_types()
            policy_check.get_metadef_objects()
            policy_check.get_metadef_properties()
            policy_check.get_metadef_tags()

            namespace_detail = Namespace.to_wsme_model(
                namespace_obj,
                get_namespace_href(namespace_obj),
                self.ns_schema_link)
            ns_filters = dict()
            ns_filters['namespace'] = namespace

            # Get objects
            object_repo = self.gateway.get_metadef_object_repo(req.context)
            db_metaobject_list = object_repo.list(filters=ns_filters)
            object_list = [MetadefObject.to_wsme_model(
                db_metaobject,
                get_object_href(namespace, db_metaobject),
                self.obj_schema_link) for db_metaobject in db_metaobject_list]
            if object_list:
                namespace_detail.objects = object_list

            # Get resource type associations
            rs_repo = self.gateway.get_metadef_resource_type_repo(
                req.context)
            db_resource_type_list = rs_repo.list(filters=ns_filters)
            resource_type_list = [ResourceTypeAssociation.to_wsme_model(
                resource_type) for resource_type in db_resource_type_list]
            if resource_type_list:
                namespace_detail.resource_type_associations = (
                    resource_type_list)

            # Get properties
            prop_repo = self.gateway.get_metadef_property_repo(req.context)
            db_properties = prop_repo.list(filters=ns_filters)
            property_list = Namespace.to_model_properties(db_properties)
            if property_list:
                namespace_detail.properties = property_list

            if filters and filters['resource_type']:
                namespace_detail = self._prefix_property_name(
                    namespace_detail, filters['resource_type'])

            # Get tags
            tag_repo = self.gateway.get_metadef_tag_repo(req.context)
            db_metatag_list = tag_repo.list(filters=ns_filters)
            tag_list = [MetadefTag(**{'name': db_metatag.name})
                        for db_metatag in db_metatag_list]
            if tag_list:
                namespace_detail.tags = tag_list

        except exception.Forbidden as e:
            LOG.debug("User not permitted to show metadata namespace "
                      "'%s'", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        return namespace_detail

    def update(self, req, user_ns, namespace):
        namespace_repo = self.gateway.get_metadef_namespace_repo(req.context)
        try:
            ns_obj = namespace_repo.get(namespace)
        except (exception.Forbidden, exception.NotFound):
            # NOTE (abhishekk): Returning 404 Not Found as the
            # namespace is outside of this user's project
            msg = _("Namespace %s not found") % namespace
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            # NOTE(abhishekk): Here we are just checking if use is authorized
            # to modify the namespace or not
            api_policy.MetadefAPIPolicy(
                req.context,
                md_resource=ns_obj,
                enforcer=self.policy).modify_metadef_namespace()

            ns_obj._old_namespace = ns_obj.namespace
            ns_obj.namespace = wsme_utils._get_value(user_ns.namespace)
            ns_obj.display_name = wsme_utils._get_value(user_ns.display_name)
            ns_obj.description = wsme_utils._get_value(user_ns.description)
            # Following optional fields will default to same values as in
            # create namespace if not specified
            ns_obj.visibility = (
                wsme_utils._get_value(user_ns.visibility) or 'private')
            ns_obj.protected = (
                wsme_utils._get_value(user_ns.protected) or False)
            ns_obj.owner = (
                wsme_utils._get_value(user_ns.owner) or req.context.owner)
            updated_namespace = namespace_repo.save(ns_obj)
        except exception.Invalid as e:
            msg = (_("Couldn't update metadata namespace: %s") % e)
            raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to update metadata namespace "
                      "'%s'", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Duplicate as e:
            raise webob.exc.HTTPConflict(explanation=e.msg)

        return Namespace.to_wsme_model(updated_namespace,
                                       get_namespace_href(updated_namespace),
                                       self.ns_schema_link)

    def delete(self, req, namespace):
        namespace_repo = self.gateway.get_metadef_namespace_repo(req.context)
        try:
            namespace_obj = namespace_repo.get(namespace)
        except (exception.Forbidden, exception.NotFound):
            # NOTE (abhishekk): Returning 404 Not Found as the
            # namespace is outside of this user's project
            msg = _("Namespace %s not found") % namespace
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            # NOTE(abhishekk): Here we are just checking user is authorized to
            # delete the namespace or not.
            api_policy.MetadefAPIPolicy(
                req.context,
                md_resource=namespace_obj,
                enforcer=self.policy).delete_metadef_namespace()
            namespace_obj.delete()
            namespace_repo.remove(namespace_obj)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to delete metadata namespace "
                      "'%s'", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

    def delete_objects(self, req, namespace):
        ns_repo = self.gateway.get_metadef_namespace_repo(req.context)
        try:
            namespace_obj = ns_repo.get(namespace)
        except (exception.Forbidden, exception.NotFound):
            # NOTE (abhishekk): Returning 404 Not Found as the
            # namespace is outside of this user's project
            msg = _("Namespace %s not found") % namespace
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            # NOTE(abhishekk): This call currently checks whether user
            # has permission to delete the namespace or not before deleting
            # the objects associated with it.
            api_policy.MetadefAPIPolicy(
                req.context,
                md_resource=namespace_obj,
                enforcer=self.policy).delete_metadef_namespace()

            namespace_obj.delete()
            ns_repo.remove_objects(namespace_obj)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to delete metadata objects "
                      "within '%s' namespace", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

    def delete_tags(self, req, namespace):
        ns_repo = self.gateway.get_metadef_namespace_repo(req.context)
        try:
            namespace_obj = ns_repo.get(namespace)
        except (exception.Forbidden, exception.NotFound):
            # NOTE (abhishekk): Returning 404 Not Found as the
            # namespace is outside of this user's project
            msg = _("Namespace %s not found") % namespace
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            # NOTE(abhishekk): This call currently checks whether user
            # has permission to delete the namespace or not before deleting
            # the objects associated with it.
            policy_check = api_policy.MetadefAPIPolicy(
                req.context,
                md_resource=namespace_obj,
                enforcer=self.policy)
            policy_check.delete_metadef_namespace()
            # NOTE(abhishekk): This call checks whether user
            # has permission to delete the tags or not.
            policy_check.delete_metadef_tags()

            namespace_obj.delete()
            ns_repo.remove_tags(namespace_obj)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to delete metadata tags "
                      "within '%s' namespace", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

    def delete_properties(self, req, namespace):
        ns_repo = self.gateway.get_metadef_namespace_repo(req.context)
        try:
            namespace_obj = ns_repo.get(namespace)
        except (exception.Forbidden, exception.NotFound):
            # NOTE (abhishekk): Returning 404 Not Found as the
            # namespace is outside of this user's project
            msg = _("Namespace %s not found") % namespace
            raise webob.exc.HTTPNotFound(explanation=msg)
        try:
            # NOTE(abhishekk): This call currently checks whether user
            # has permission to delete the namespace or not before deleting
            # the objects associated with it.
            api_policy.MetadefAPIPolicy(
                req.context,
                md_resource=namespace_obj,
                enforcer=self.policy).delete_metadef_namespace()

            namespace_obj.delete()
            ns_repo.remove_properties(namespace_obj)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to delete metadata properties "
                      "within '%s' namespace", namespace)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)

    def _prefix_property_name(self, namespace_detail, user_resource_type):
        prefix = None
        if user_resource_type and namespace_detail.resource_type_associations:
            for resource_type in namespace_detail.resource_type_associations:
                if resource_type.name == user_resource_type:
                    prefix = resource_type.prefix
                    break

        if prefix:
            if namespace_detail.properties:
                new_property_dict = dict()
                for (key, value) in namespace_detail.properties.items():
                    new_property_dict[prefix + key] = value
                namespace_detail.properties = new_property_dict

            if namespace_detail.objects:
                for object in namespace_detail.objects:
                    new_object_property_dict = dict()
                    for (key, value) in object.properties.items():
                        new_object_property_dict[prefix + key] = value
                    object.properties = new_object_property_dict

                    if object.required and len(object.required) > 0:
                        required = [prefix + name for name in object.required]
                        object.required = required

        return namespace_detail


class RequestDeserializer(wsgi.JSONRequestDeserializer):
    _disallowed_properties = ['self', 'schema', 'created_at', 'updated_at']

    def __init__(self, schema=None):
        super(RequestDeserializer, self).__init__()
        self.schema = schema or get_schema()

    def _get_request_body(self, request):
        output = super(RequestDeserializer, self).default(request)
        if 'body' not in output:
            msg = _('Body expected in request.')
            raise webob.exc.HTTPBadRequest(explanation=msg)
        return output['body']

    @classmethod
    def _check_allowed(cls, image):
        for key in cls._disallowed_properties:
            if key in image:
                msg = _("Attribute '%s' is read-only.") % key
                raise webob.exc.HTTPForbidden(explanation=msg)

    def index(self, request):
        params = request.params.copy()
        limit = params.pop('limit', None)
        marker = params.pop('marker', None)
        sort_dir = params.pop('sort_dir', 'desc')

        if limit is None:
            limit = CONF.limit_param_default
        limit = min(CONF.api_limit_max, int(limit))

        query_params = {
            'sort_key': params.pop('sort_key', 'created_at'),
            'sort_dir': self._validate_sort_dir(sort_dir),
            'filters': self._get_filters(params)
        }

        if marker is not None:
            query_params['marker'] = marker

        if limit is not None:
            query_params['limit'] = self._validate_limit(limit)

        return query_params

    def _validate_sort_dir(self, sort_dir):
        if sort_dir not in ['asc', 'desc']:
            msg = _('Invalid sort direction: %s') % sort_dir
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return sort_dir

    def _get_filters(self, filters):
        visibility = filters.get('visibility')
        if visibility:
            if visibility not in ['public', 'private']:
                msg = _('Invalid visibility value: %s') % visibility
                raise webob.exc.HTTPBadRequest(explanation=msg)

        return filters

    def _validate_limit(self, limit):
        try:
            limit = int(limit)
        except ValueError:
            msg = _("limit param must be an integer")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        if limit < 0:
            msg = _("limit param must be positive")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        return limit

    def show(self, request):
        params = request.params.copy()
        query_params = {
            'filters': self._get_filters(params)
        }
        return query_params

    def create(self, request):
        body = self._get_request_body(request)
        self._check_allowed(body)
        try:
            self.schema.validate(body)
        except exception.InvalidObject as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        namespace = json.fromjson(Namespace, body)
        return dict(namespace=namespace)

    def update(self, request):
        body = self._get_request_body(request)
        self._check_allowed(body)
        try:
            self.schema.validate(body)
        except exception.InvalidObject as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)
        namespace = json.fromjson(Namespace, body)
        return dict(user_ns=namespace)


class ResponseSerializer(wsgi.JSONResponseSerializer):
    def __init__(self, schema=None):
        super(ResponseSerializer, self).__init__()
        self.schema = schema

    def create(self, response, namespace):
        ns_json = json.tojson(Namespace, namespace)
        response = self.__render(ns_json, response, http.CREATED)
        response.location = get_namespace_href(namespace)

    def show(self, response, namespace):
        ns_json = json.tojson(Namespace, namespace)
        response = self.__render(ns_json, response)

    def index(self, response, result):
        params = dict(response.request.params)
        params.pop('marker', None)
        query = urlparse.urlencode(params)
        result.first = "/v2/metadefs/namespaces"
        result.schema = "/v2/schemas/metadefs/namespaces"
        if query:
            result.first = '%s?%s' % (result.first, query)
        if result.next:
            params['marker'] = result.next
            next_query = urlparse.urlencode(params)
            result.next = '/v2/metadefs/namespaces?%s' % next_query

        ns_json = json.tojson(Namespaces, result)
        response = self.__render(ns_json, response)

    def update(self, response, namespace):
        ns_json = json.tojson(Namespace, namespace)
        response = self.__render(ns_json, response, http.OK)

    def delete(self, response, result):
        response.status_int = http.NO_CONTENT

    def delete_objects(self, response, result):
        response.status_int = http.NO_CONTENT

    def delete_properties(self, response, result):
        response.status_int = http.NO_CONTENT

    def delete_tags(self, response, result):
        response.status_int = http.NO_CONTENT

    def __render(self, json_data, response, response_status=None):
        body = jsonutils.dumps(json_data, ensure_ascii=False)
        response.unicode_body = body
        response.content_type = 'application/json'
        if response_status:
            response.status_int = response_status
        return response


def _get_base_definitions():
    return get_schema_definitions()


def get_schema_definitions():
    return {
        "positiveInteger": {
            "type": "integer",
            "minimum": 0
        },
        "positiveIntegerDefault0": {
            "allOf": [
                {"$ref": "#/definitions/positiveInteger"},
                {"default": 0}
            ]
        },
        "stringArray": {
            "type": "array",
            "items": {"type": "string"},
            # "minItems": 1,
            "uniqueItems": True
        },
        "property": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "required": ["title", "type"],
                "properties": {
                    "name": {
                        "type": "string",
                        "maxLength": 80
                    },
                    "title": {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    },
                    "operators": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "type": {
                        "type": "string",
                        "enum": [
                            "array",
                            "boolean",
                            "integer",
                            "number",
                            "object",
                            "string",
                            None
                        ]
                    },
                    "required": {
                        "$ref": "#/definitions/stringArray"
                    },
                    "minimum": {
                        "type": "number"
                    },
                    "maximum": {
                        "type": "number"
                    },
                    "maxLength": {
                        "$ref": "#/definitions/positiveInteger"
                    },
                    "minLength": {
                        "$ref": "#/definitions/positiveIntegerDefault0"
                    },
                    "pattern": {
                        "type": "string",
                        "format": "regex"
                    },
                    "enum": {
                        "type": "array"
                    },
                    "readonly": {
                        "type": "boolean"
                    },
                    "default": {},
                    "items": {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string",
                                "enum": [
                                    "array",
                                    "boolean",
                                    "integer",
                                    "number",
                                    "object",
                                    "string",
                                    None
                                ]
                            },
                            "enum": {
                                "type": "array"
                            }
                        }
                    },
                    "maxItems": {
                        "$ref": "#/definitions/positiveInteger"
                    },
                    "minItems": {
                        "$ref": "#/definitions/positiveIntegerDefault0"
                    },
                    "uniqueItems": {
                        "type": "boolean",
                        "default": False
                    },
                    "additionalItems": {
                        "type": "boolean"
                    },
                }
            }
        }
    }


def _get_base_properties():
    return {
        "namespace": {
            "type": "string",
            "description": _("The unique namespace text."),
            "maxLength": 80,
        },
        "display_name": {
            "type": "string",
            "description": _("The user friendly name for the namespace. Used "
                             "by UI if available."),
            "maxLength": 80,
        },
        "description": {
            "type": "string",
            "description": _("Provides a user friendly description of the "
                             "namespace."),
            "maxLength": 500,
        },
        "visibility": {
            "type": "string",
            "description": _("Scope of namespace accessibility."),
            "enum": ["public", "private"],
        },
        "protected": {
            "type": "boolean",
            "description": _("If true, namespace will not be deletable."),
        },
        "owner": {
            "type": "string",
            "description": _("Owner of the namespace."),
            "maxLength": 255,
        },
        "created_at": {
            "type": "string",
            "readOnly": True,
            "description": _("Date and time of namespace creation"),
            "format": "date-time"
        },
        "updated_at": {
            "type": "string",
            "readOnly": True,
            "description": _("Date and time of the last namespace"
                             " modification"),
            "format": "date-time"
        },
        "schema": {
            'readOnly': True,
            "type": "string"
        },
        "self": {
            'readOnly': True,
            "type": "string"
        },
        "resource_type_associations": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "prefix": {
                        "type": "string"
                    },
                    "properties_target": {
                        "type": "string"
                    }
                }
            }
        },
        "properties": {
            "$ref": "#/definitions/property"
        },
        "objects": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    },
                    "required": {
                        "$ref": "#/definitions/stringArray"
                    },
                    "properties": {
                        "$ref": "#/definitions/property"
                    },
                }
            }
        },
        "tags": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    }
                }
            }
        },
    }


def get_schema():
    properties = _get_base_properties()
    definitions = _get_base_definitions()
    mandatory_attrs = Namespace.get_mandatory_attrs()
    schema = glance.schema.Schema(
        'namespace',
        properties,
        required=mandatory_attrs,
        definitions=definitions
    )
    return schema


def get_collection_schema():
    namespace_schema = get_schema()
    return glance.schema.CollectionSchema('namespaces', namespace_schema)


def get_namespace_href(namespace):
    base_href = '/v2/metadefs/namespaces/%s' % namespace.namespace
    return base_href


def get_object_href(namespace_name, metadef_object):
    base_href = ('/v2/metadefs/namespaces/%s/objects/%s' %
                 (namespace_name, metadef_object.name))
    return base_href


def get_tag_href(namespace_name, metadef_tag):
    base_href = ('/v2/metadefs/namespaces/%s/tags/%s' %
                 (namespace_name, metadef_tag.name))
    return base_href


def create_resource():
    """Namespaces resource factory method"""
    schema = get_schema()
    deserializer = RequestDeserializer(schema)
    serializer = ResponseSerializer(schema)
    controller = NamespaceController()
    return wsgi.Resource(controller, deserializer, serializer)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
import datetime
import errno
import hashlib
import io
import os
import tempfile
import time
from unittest import mock

import fixtures
import glance_store as store
from oslo_config import cfg
from oslo_utils import fileutils
from oslo_utils import units

from glance import async_
from glance.common import exception
from glance import context
from glance import gateway as glance_gateway
from glance import image_cache
from glance.image_cache import prefetcher
from glance.tests.unit import utils as unit_test_utils
from glance.tests import utils as test_utils
from glance.tests.utils import skip_if_disabled
from glance.tests.utils import xattr_writes_supported

FIXTURE_LENGTH = 1024
FIXTURE_DATA = b'*' * FIXTURE_LENGTH
CONF = cfg.CONF


class ImageCacheTestCase(object):

    def _setup_fixture_file(self):
        FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)

        self.assertFalse(self.cache.is_cached(1))

        self.assertTrue(self.cache.cache_image_file(1, FIXTURE_FILE))

        self.assertTrue(self.cache.is_cached(1))

    @skip_if_disabled
    def test_is_cached(self):
        """Verify is_cached(1) returns 0, then add something to the cache
        and verify is_cached(1) returns 1.
        """
        self._setup_fixture_file()

    @skip_if_disabled
    def test_read(self):
        """Verify is_cached(1) returns 0, then add something to the cache
        and verify after a subsequent read from the cache that
        is_cached(1) returns 1.
        """
        self._setup_fixture_file()

        buff = io.BytesIO()
        with self.cache.open_for_read(1) as cache_file:
            for chunk in cache_file:
                buff.write(chunk)

        self.assertEqual(FIXTURE_DATA, buff.getvalue())

    @skip_if_disabled
    def test_open_for_read(self):
        """Test convenience wrapper for opening a cache file via
        its image identifier.
        """
        self._setup_fixture_file()

        buff = io.BytesIO()
        with self.cache.open_for_read(1) as cache_file:
            for chunk in cache_file:
                buff.write(chunk)

        self.assertEqual(FIXTURE_DATA, buff.getvalue())

    @skip_if_disabled
    def test_get_image_size(self):
        """Test convenience wrapper for querying cache file size via
        its image identifier.
        """
        self._setup_fixture_file()

        size = self.cache.get_image_size(1)

        self.assertEqual(FIXTURE_LENGTH, size)

    @skip_if_disabled
    def test_delete(self):
        """Test delete method that removes an image from the cache."""
        self._setup_fixture_file()

        self.cache.delete_cached_image(1)

        self.assertFalse(self.cache.is_cached(1))

    @skip_if_disabled
    def test_delete_all(self):
        """Test delete method that removes an image from the cache."""
        for image_id in (1, 2):
            self.assertFalse(self.cache.is_cached(image_id))

        for image_id in (1, 2):
            FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)
            self.assertTrue(self.cache.cache_image_file(image_id,
                                                        FIXTURE_FILE))

        for image_id in (1, 2):
            self.assertTrue(self.cache.is_cached(image_id))

        self.cache.delete_all_cached_images()

        for image_id in (1, 2):
            self.assertFalse(self.cache.is_cached(image_id))

    def _test_clean_invalid_path(self, failure=False):
        invalid_file_path = os.path.join(self.cache_dir, 'invalid', '1')
        invalid_file = open(invalid_file_path, 'wb')
        invalid_file.write(FIXTURE_DATA)
        invalid_file.close()

        self.assertTrue(os.path.exists(invalid_file_path))

        if failure:
            with mock.patch.object(
                    fileutils, 'delete_if_exists') as mock_delete:
                mock_delete.side_effect = OSError(errno.ENOENT, '')
                try:
                    self.cache.clean()
                except OSError:
                    self.assertTrue(os.path.exists(invalid_file_path))
        else:
            self.cache.clean()
            self.assertFalse(os.path.exists(invalid_file_path))

    @skip_if_disabled
    def test_clean_invalid_path(self):
        """Test the clean method removes expected image from invalid path."""
        self._test_clean_invalid_path()

    @skip_if_disabled
    def test_clean_stalled(self):
        """Test the clean method removes expected images."""
        incomplete_file_path = os.path.join(self.cache_dir, 'incomplete', '1')
        incomplete_file = open(incomplete_file_path, 'wb')
        incomplete_file.write(FIXTURE_DATA)
        incomplete_file.close()

        self.assertTrue(os.path.exists(incomplete_file_path))

        self.cache.clean(stall_time=0)

        self.assertFalse(os.path.exists(incomplete_file_path))

    def _test_clean_stall_time(self, stall_time=None, days=2,
                               stall_failed=False):
        """
        Test the clean method removes the stalled images as expected
        """
        incomplete_file_path_1 = os.path.join(self.cache_dir,
                                              'incomplete', '1')
        incomplete_file_path_2 = os.path.join(self.cache_dir,
                                              'incomplete', '2')
        for f in (incomplete_file_path_1, incomplete_file_path_2):
            incomplete_file = open(f, 'wb')
            incomplete_file.write(FIXTURE_DATA)
            incomplete_file.close()

        mtime = os.path.getmtime(incomplete_file_path_1)
        pastday = (datetime.datetime.fromtimestamp(mtime) -
                   datetime.timedelta(days=days))
        atime = int(time.mktime(pastday.timetuple()))
        mtime = atime
        os.utime(incomplete_file_path_1, (atime, mtime))

        self.assertTrue(os.path.exists(incomplete_file_path_1))
        self.assertTrue(os.path.exists(incomplete_file_path_2))

        # If stall_time is None then it will wait for default time
        # of `image_cache_stall_time` which is 24 hours
        if stall_failed:
            with mock.patch.object(
                    fileutils, 'delete_if_exists') as mock_delete:
                mock_delete.side_effect = OSError(errno.ENOENT, '')
                self.cache.clean(stall_time=stall_time)
                self.assertTrue(os.path.exists(incomplete_file_path_1))
        else:
            self.cache.clean(stall_time=stall_time)
            self.assertFalse(os.path.exists(incomplete_file_path_1))

        self.assertTrue(os.path.exists(incomplete_file_path_2))

    @skip_if_disabled
    def test_clean_stalled_none_stall_time(self):
        self._test_clean_stall_time()

    @skip_if_disabled
    def test_clean_stalled_nonzero_stall_time(self):
        """Test the clean method removes expected images."""
        self._test_clean_stall_time(stall_time=3600, days=1)

    @skip_if_disabled
    def test_prune(self):
        """
        Test that pruning the cache works as expected...
        """
        self.assertEqual(0, self.cache.get_cache_size())

        # Add a bunch of images to the cache. The max cache size for the cache
        # is set to 5KB and each image is 1K. We use 11 images in this test.
        # The first 10 are added to and retrieved from cache in the same order.
        # Then, the 11th image is added to cache but not retrieved before we
        # prune. We should see only 5 images left after pruning, and the
        # images that are least recently accessed should be the ones pruned...
        for x in range(10):
            FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)
            self.assertTrue(self.cache.cache_image_file(x, FIXTURE_FILE))

        self.assertEqual(10 * units.Ki, self.cache.get_cache_size())

        # OK, hit the images that are now cached...
        for x in range(10):
            buff = io.BytesIO()
            with self.cache.open_for_read(x) as cache_file:
                for chunk in cache_file:
                    buff.write(chunk)

        # Add a new image to cache.
        # This is specifically to test the bug: 1438564
        FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)
        self.assertTrue(self.cache.cache_image_file(99, FIXTURE_FILE))

        self.cache.prune()

        self.assertEqual(5 * units.Ki, self.cache.get_cache_size())

        # Ensure images 0, 1, 2, 3, 4 & 5 are not cached anymore
        for x in range(0, 6):
            self.assertFalse(self.cache.is_cached(x),
                             "Image %s was cached!" % x)

        # Ensure images 6, 7, 8 and 9 are still cached
        for x in range(6, 10):
            self.assertTrue(self.cache.is_cached(x),
                            "Image %s was not cached!" % x)

        # Ensure the newly added image, 99, is still cached
        self.assertTrue(self.cache.is_cached(99), "Image 99 was not cached!")

    @skip_if_disabled
    def test_prune_to_zero(self):
        """Test that an image_cache_max_size of 0 doesn't kill the pruner

        This is a test specifically for LP #1039854
        """
        self.assertEqual(0, self.cache.get_cache_size())

        FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)
        self.assertTrue(self.cache.cache_image_file('xxx', FIXTURE_FILE))

        self.assertEqual(1024, self.cache.get_cache_size())

        # OK, hit the image that is now cached...
        buff = io.BytesIO()
        with self.cache.open_for_read('xxx') as cache_file:
            for chunk in cache_file:
                buff.write(chunk)

        self.config(image_cache_max_size=0)
        self.cache.prune()

        self.assertEqual(0, self.cache.get_cache_size())
        self.assertFalse(self.cache.is_cached('xxx'))

    @skip_if_disabled
    def test_queue(self):
        """
        Test that queueing works properly
        """

        self.assertFalse(self.cache.is_cached(1))
        self.assertFalse(self.cache.is_queued(1))

        FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)

        self.assertTrue(self.cache.queue_image(1))

        self.assertTrue(self.cache.is_queued(1))
        self.assertFalse(self.cache.is_cached(1))

        # Should not return True if the image is already
        # queued for caching...
        self.assertFalse(self.cache.queue_image(1))

        self.assertFalse(self.cache.is_cached(1))

        # Test that we return False if we try to queue
        # an image that has already been cached

        self.assertTrue(self.cache.cache_image_file(1, FIXTURE_FILE))

        self.assertFalse(self.cache.is_queued(1))
        self.assertTrue(self.cache.is_cached(1))

        self.assertFalse(self.cache.queue_image(1))

        self.cache.delete_cached_image(1)

        # Test that we return false if image is being cached
        incomplete_file_path = os.path.join(self.cache_dir, 'incomplete', '1')
        incomplete_file = open(incomplete_file_path, 'wb')
        incomplete_file.write(FIXTURE_DATA)
        incomplete_file.close()

        self.assertFalse(self.cache.is_queued(1))
        self.assertFalse(self.cache.is_cached(1))
        self.assertTrue(self.cache.driver.is_being_cached(1))

        self.assertFalse(self.cache.queue_image(1))
        self.cache.clean(stall_time=0)

        for x in range(3):
            self.assertTrue(self.cache.queue_image(x))

        self.assertEqual(['0', '1', '2'],
                         self.cache.get_queued_images())

    @skip_if_disabled
    def test_open_for_write_good(self):
        """
        Test to see if open_for_write works in normal case
        """

        # test a good case
        image_id = '1'
        self.assertFalse(self.cache.is_cached(image_id))
        with self.cache.driver.open_for_write(image_id) as cache_file:
            cache_file.write(b'a')
        self.assertTrue(self.cache.is_cached(image_id),
                        "Image %s was NOT cached!" % image_id)
        # make sure it has tidied up
        incomplete_file_path = os.path.join(self.cache_dir,
                                            'incomplete', image_id)
        invalid_file_path = os.path.join(self.cache_dir, 'invalid', image_id)
        self.assertFalse(os.path.exists(incomplete_file_path))
        self.assertFalse(os.path.exists(invalid_file_path))

    @skip_if_disabled
    def test_open_for_write_with_exception(self):
        """
        Test to see if open_for_write works in a failure case for each driver
        This case is where an exception is raised while the file is being
        written. The image is partially filled in cache and filling won't
        resume so verify the image is moved to invalid/ directory
        """
        # test a case where an exception is raised while the file is open
        image_id = '1'
        self.assertFalse(self.cache.is_cached(image_id))
        try:
            with self.cache.driver.open_for_write(image_id):
                raise IOError
        except Exception as e:
            self.assertIsInstance(e, IOError)
        self.assertFalse(self.cache.is_cached(image_id),
                         "Image %s was cached!" % image_id)
        # make sure it has tidied up
        incomplete_file_path = os.path.join(self.cache_dir,
                                            'incomplete', image_id)
        invalid_file_path = os.path.join(self.cache_dir, 'invalid', image_id)
        self.assertFalse(os.path.exists(incomplete_file_path))
        self.assertTrue(os.path.exists(invalid_file_path))

    @skip_if_disabled
    def test_caching_iterator(self):
        """
        Test to see if the caching iterator interacts properly with the driver
        When the iterator completes going through the data the driver should
        have closed the image and placed it correctly
        """
        # test a case where an exception NOT raised while the file is open,
        # and a consuming iterator completes
        def consume(image_id):
            data = [b'a', b'b', b'c', b'd', b'e', b'f']
            checksum = None
            caching_iter = self.cache.get_caching_iter(image_id, checksum,
                                                       iter(data))
            self.assertEqual(data, list(caching_iter))

        image_id = '1'
        self.assertFalse(self.cache.is_cached(image_id))
        consume(image_id)
        self.assertTrue(self.cache.is_cached(image_id),
                        "Image %s was NOT cached!" % image_id)
        # make sure it has tidied up
        incomplete_file_path = os.path.join(self.cache_dir,
                                            'incomplete', image_id)
        invalid_file_path = os.path.join(self.cache_dir, 'invalid', image_id)
        self.assertFalse(os.path.exists(incomplete_file_path))
        self.assertFalse(os.path.exists(invalid_file_path))

    @skip_if_disabled
    def test_caching_iterator_handles_backend_failure(self):
        """
        Test that when the backend fails, caching_iter does not continue trying
        to consume data, and rolls back the cache.
        """
        def faulty_backend():
            data = [b'a', b'b', b'c', b'Fail', b'd', b'e', b'f']
            for d in data:
                if d == b'Fail':
                    raise exception.GlanceException('Backend failure')
                yield d

        def consume(image_id):
            caching_iter = self.cache.get_caching_iter(image_id, None,
                                                       faulty_backend())
            # exercise the caching_iter
            list(caching_iter)

        image_id = '1'
        self.assertRaises(exception.GlanceException, consume, image_id)
        # make sure bad image was not cached
        self.assertFalse(self.cache.is_cached(image_id))

    @skip_if_disabled
    def test_caching_iterator_falloffend(self):
        """
        Test to see if the caching iterator interacts properly with the driver
        in a case where the iterator is only partially consumed. In this case
        the image is only partially filled in cache and filling won't resume.
        When the iterator goes out of scope the driver should have closed the
        image and moved it from incomplete/ to invalid/
        """
        # test a case where a consuming iterator just stops.
        def falloffend(image_id):
            data = [b'a', b'b', b'c', b'd', b'e', b'f']
            checksum = None
            caching_iter = self.cache.get_caching_iter(image_id, checksum,
                                                       iter(data))
            self.assertEqual(b'a', next(caching_iter))

        image_id = '1'
        self.assertFalse(self.cache.is_cached(image_id))
        falloffend(image_id)
        self.assertFalse(self.cache.is_cached(image_id),
                         "Image %s was cached!" % image_id)
        # make sure it has tidied up
        incomplete_file_path = os.path.join(self.cache_dir,
                                            'incomplete', image_id)
        invalid_file_path = os.path.join(self.cache_dir, 'invalid', image_id)
        self.assertFalse(os.path.exists(incomplete_file_path))
        self.assertTrue(os.path.exists(invalid_file_path))

    @skip_if_disabled
    def test_gate_caching_iter_good_checksum(self):
        image = b"12345678990abcdefghijklmnop"
        image_id = 123

        md5 = hashlib.md5(usedforsecurity=False)
        md5.update(image)
        checksum = md5.hexdigest()

        with mock.patch('glance.db.get_api') as mock_get_db:
            db = unit_test_utils.FakeDB(initialize=False)
            mock_get_db.return_value = db
            cache = image_cache.ImageCache()

        img_iter = cache.get_caching_iter(image_id, checksum, [image])
        for chunk in img_iter:
            pass
        # checksum is valid, fake image should be cached:
        self.assertTrue(cache.is_cached(image_id))

    @skip_if_disabled
    def test_gate_caching_iter_bad_checksum(self):
        image = b"12345678990abcdefghijklmnop"
        image_id = 123
        checksum = "foobar"  # bad.

        with mock.patch('glance.db.get_api') as mock_get_db:
            db = unit_test_utils.FakeDB(initialize=False)
            mock_get_db.return_value = db
            cache = image_cache.ImageCache()

        img_iter = cache.get_caching_iter(image_id, checksum, [image])

        def reader():
            for chunk in img_iter:
                pass
        self.assertRaises(exception.GlanceException, reader)
        # checksum is invalid, caching will fail:
        self.assertFalse(cache.is_cached(image_id))


class TestImageCacheXattr(test_utils.BaseTestCase,
                          ImageCacheTestCase):

    """Tests image caching when xattr is used in cache"""

    def setUp(self):
        """
        Test to see if the pre-requisites for the image cache
        are working (python-xattr installed and xattr support on the
        filesystem)
        """
        super(TestImageCacheXattr, self).setUp()

        if getattr(self, 'disable', False):
            return

        self.cache_dir = self.useFixture(fixtures.TempDir()).path

        if not getattr(self, 'inited', False):
            try:
                import xattr  # noqa
            except ImportError:
                self.inited = True
                self.disabled = True
                self.disabled_message = ("python-xattr not installed.")
                return

        self.inited = True
        self.disabled = False
        self.config(image_cache_dir=self.cache_dir,
                    image_cache_driver='xattr',
                    image_cache_max_size=5 * units.Ki)
        self.cache = image_cache.ImageCache()

        if not xattr_writes_supported(self.cache_dir):
            self.inited = True
            self.disabled = True
            self.disabled_message = ("filesystem does not support xattr")
            return


class TestImageCacheCentralizedDb(test_utils.BaseTestCase,
                                  ImageCacheTestCase):

    """Tests image caching when Centralized DB is used in cache"""

    def setUp(self):
        super(TestImageCacheCentralizedDb, self).setUp()

        self.inited = True
        self.disabled = False
        self.cache_dir = self.useFixture(fixtures.TempDir()).path
        self.config(image_cache_dir=self.cache_dir,
                    image_cache_driver='centralized_db',
                    image_cache_max_size=5 * units.Ki,
                    worker_self_reference_url='http://workerx')

        with mock.patch('glance.db.get_api') as mock_get_db:
            self.db = unit_test_utils.FakeDB(initialize=False)
            mock_get_db.return_value = self.db
            self.cache = image_cache.ImageCache()

    def test_node_reference_create_duplicate(self):
        with mock.patch('glance.db.get_api') as mock_get_db:
            self.db = unit_test_utils.FakeDB(initialize=False)
            mock_get_db.return_value = self.db
            with mock.patch.object(
                    self.db, 'node_reference_create') as mock_node_create:
                mock_node_create.side_effect = exception.Duplicate
                with mock.patch.object(
                        image_cache.drivers.centralized_db, 'LOG') as mock_log:
                    image_cache.ImageCache()
                    expected_calls = [
                        mock.call('Node reference is already recorded, '
                                  'ignoring it')
                    ]
                    mock_log.debug.assert_has_calls(expected_calls)

    def test_get_least_recently_accessed_os_error(self):
        self.assertEqual(0, self.cache.get_cache_size())
        for x in range(10):
            FIXTURE_FILE = io.BytesIO(FIXTURE_DATA)
            self.assertTrue(self.cache.cache_image_file(x, FIXTURE_FILE))

        self.assertEqual(10 * units.Ki, self.cache.get_cache_size())

        with mock.patch.object(os, 'stat') as mock_stat:
            mock_stat.side_effect = OSError
            image_id, size = self.cache.driver.get_least_recently_accessed()
            self.assertEqual(0, size)

    @skip_if_disabled
    def test_clean_stalled_fails(self):
        """Test the clean method fails to delete file, ignores the failure"""
        self._test_clean_stall_time(stall_time=3600, days=1,
                                    stall_failed=True)

    @skip_if_disabled
    def test_clean_invalid_path_fails(self):
        """Test the clean method fails to remove image from invalid path."""
        self._test_clean_invalid_path(failure=True)


class TestImageCacheSqlite(test_utils.BaseTestCase,
                           ImageCacheTestCase):

    """Tests image caching when SQLite is used in cache"""

    def setUp(self):
        super(TestImageCacheSqlite, self).setUp()

        self.cache_dir = self.useFixture(fixtures.TempDir()).path
        self.config(image_cache_dir=self.cache_dir,
                    image_cache_driver='sqlite',
                    image_cache_max_size=5 * units.Ki)
        self.cache = image_cache.ImageCache()

    @mock.patch('glance.db.get_api')
    def _test_prefetcher(self, mock_get_db):
        self.config(enabled_backends={'cheap': 'file'})
        store.register_store_opts(CONF)
        self.config(filesystem_store_datadir='/tmp', group='cheap')
        store.create_multi_stores(CONF)

        tempf = tempfile.NamedTemporaryFile()
        tempf.write(b'foo')

        db = unit_test_utils.FakeDB(initialize=False)
        mock_get_db.return_value = db

        ctx = context.RequestContext(is_admin=True, roles=['admin'])
        gateway = glance_gateway.Gateway()
        image_factory = gateway.get_image_factory(ctx)
        image_repo = gateway.get_repo(ctx)
        fetcher = prefetcher.Prefetcher()

        # Create an image with no values set and queue it
        image = image_factory.new_image()
        image_repo.add(image)
        fetcher.cache.queue_image(image.image_id)

        # Image is not active, so it should fail to cache, but remain queued
        self.assertFalse(fetcher.run())
        self.assertFalse(fetcher.cache.is_cached(image.image_id))
        self.assertTrue(fetcher.cache.is_queued(image.image_id))

        # Set the disk/container format and give it a location
        image.disk_format = 'raw'
        image.container_format = 'bare'
        image.status = 'active'
        loc = {'url': 'file://%s' % tempf.name, 'metadata': {'store': 'cheap'}}
        with mock.patch('glance.location._check_image_location'):
            # FIXME(danms): Why do I have to do this?
            image.locations = [loc]
        image_repo.save(image)

        # Image is now active and has a location, so it should cache
        self.assertTrue(fetcher.run())
        self.assertTrue(fetcher.cache.is_cached(image.image_id))
        self.assertFalse(fetcher.cache.is_queued(image.image_id))

    @mock.patch('glance.async_._THREADPOOL_MODEL', new=None)
    def test_prefetcher_greenthread(self):
        async_.set_threadpool_model('eventlet')
        self._test_prefetcher()

    @mock.patch('glance.async_._THREADPOOL_MODEL', new=None)
    def test_prefetcher_native(self):
        async_.set_threadpool_model('native')
        self._test_prefetcher()


class TestImageCacheNoDep(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageCacheNoDep, self).setUp()

        self.driver = None

        def init_driver(self2):
            self2.driver = self.driver

        self.mock_object(image_cache.ImageCache, 'init_driver', init_driver)

    def test_get_caching_iter_when_write_fails(self):

        class FailingFile(object):

            def write(self, data):
                if data == "Fail":
                    raise IOError

        class FailingFileDriver(object):

            def is_cacheable(self, *args, **kwargs):
                return True

            @contextmanager
            def open_for_write(self, *args, **kwargs):
                yield FailingFile()

        self.driver = FailingFileDriver()
        cache = image_cache.ImageCache()
        data = [b'a', b'b', b'c', b'Fail', b'd', b'e', b'f']

        caching_iter = cache.get_caching_iter('dummy_id', None, iter(data))
        self.assertEqual(data, list(caching_iter))

    def test_get_caching_iter_when_open_fails(self):

        class OpenFailingDriver(object):

            def is_cacheable(self, *args, **kwargs):
                return True

            @contextmanager
            def open_for_write(self, *args, **kwargs):
                raise IOError

        self.driver = OpenFailingDriver()
        cache = image_cache.ImageCache()
        data = [b'a', b'b', b'c', b'd', b'e', b'f']

        caching_iter = cache.get_caching_iter('dummy_id', None, iter(data))
        self.assertEqual(data, list(caching_iter))


class TestImagePrefetcher(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImagePrefetcher, self).setUp()
        self.cache_dir = self.useFixture(fixtures.TempDir()).path
        self.config(image_cache_dir=self.cache_dir,
                    image_cache_driver='xattr',
                    image_cache_max_size=5 * units.Ki)
        self.prefetcher = prefetcher.Prefetcher()

    def test_fetch_image_into_cache_without_auth(self):
        with mock.patch.object(self.prefetcher.gateway,
                               'get_repo') as mock_get:
            self.prefetcher.fetch_image_into_cache('fake-image-id')
            mock_get.assert_called_once_with(mock.ANY)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import gzip
import hashlib
import http.client as http
import http.server as http_server
import os
import threading
import time
import zipfile

from oslo_serialization import jsonutils
import testtools

import glance.async_.flows.plugins.image_decompression as image_decompression
from glance.tests import functional
from glance.tests.functional import ft_utils as func_utils
from glance.tests.functional.v2.test_images import ImageAPIHelper


class TestImageDecompression(functional.SynchronousAPIBase):
    """Functional tests for image decompression plugin.

    Tests the image decompression plugin during image import operations
    using both glance-direct and web-download methods.
    """

    def setUp(self, single_store=True):
        super(TestImageDecompression, self).setUp(single_store=single_store)
        # Import to register image_import_opts config group otherwise tests
        # will fail with NoSuchOptGroup error.
        from glance.async_.flows import api_image_import  # noqa
        self.setup_database()
        self.setup_simple_paste()
        if single_store:
            self.setup_single_store()
        else:
            self.setup_stores()

        # Enable image decompression plugin
        self.config(image_import_plugins=['image_decompression'],
                    group='image_import_opts')
        self.config(enabled_import_methods=['glance-direct', 'web-download'])

        self.api_methods = ImageAPIHelper(
            self.api_get, self.api_post, self.api_put, self.api_delete)

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': '70b8c4b4-4e3e-4e3e-9e3e-4e3e4e3e4e3e',
            'X-Project-Id': self.TENANT,
            'X-Roles': 'member,reader',
            'content-type': 'application/json'
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def _create_gzip_file(self, content, filename='test.gz'):
        """Create a gzipped file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with gzip.open(filepath, 'wb') as f:
            f.write(content)
        return filepath

    def _create_zip_file(self, content, filename='test.zip',
                         entry_name='test_image.raw'):
        """Create a ZIP file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with zipfile.ZipFile(filepath, 'w') as zf:
            zf.writestr(entry_name, content)
        return filepath

    def _create_zip_file_multiple(self, files_dict, filename='test.zip'):
        """Create a ZIP file with multiple entries."""
        filepath = os.path.join(self.test_dir, filename)
        with zipfile.ZipFile(filepath, 'w') as zf:
            for entry_name, content in files_dict.items():
                zf.writestr(entry_name, content)
        return filepath

    def _create_uncompressed_file(self, content, filename='test.raw'):
        """Create an uncompressed file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with open(filepath, 'wb') as f:
            f.write(content)
        return filepath

    def _calculate_checksum(self, data):
        """Calculate MD5 checksum of data."""
        return hashlib.md5(data, usedforsecurity=False).hexdigest()

    def _calculate_sha512(self, data):
        """Calculate SHA512 checksum of data."""
        return hashlib.sha512(data).hexdigest()

    def _wait_for_task_failure(self, image_id, max_sec=40, delay_sec=0.2,
                               start_delay_sec=1):
        """Wait for import task to fail.

        This method checks the task status associated with the image to
        determine if the import task has failed.

        :param image_id: The image ID to check
        :param max_sec: Maximum seconds to wait (default: 40)
        :param delay_sec: Seconds to sleep between checks (default: 0.2)
        :param start_delay_sec: Seconds to wait before first check (default: 1)
        :returns: The task dict from the API response
        """
        start_time = time.time()
        done_time = start_time + max_sec
        if start_delay_sec:
            time.sleep(start_delay_sec)

        while time.time() <= done_time:
            try:
                task = self._get_latest_task(image_id)
                if task['status'] == 'failure':
                    return task
                elif task['status'] == 'success':
                    self.fail("Import unexpectedly succeeded "
                              "(task status=success)")
            except (KeyError, IndexError):
                # Task may not exist yet, continue checking
                pass

            time.sleep(delay_sec)

        # Final check - verify task status
        task = self._get_latest_task(image_id)
        self.assertEqual('failure', task['status'])
        return task

    def _start_binary_http_server(self, data):
        """Start an HTTP server serving binary data.

        Returns a tuple of (httpd, thread, uri) where:
        - httpd: The HTTP server instance (for cleanup)
        - thread: The server thread (for reference)
        - uri: The URI to use for web-download
        """
        class BinaryHTTPRequestHandler(http_server.BaseHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                self.data = data
                super().__init__(*args, **kwargs)

            def do_GET(self):
                self.send_response(200)
                self.send_header('Content-Length', str(len(self.data)))
                self.end_headers()
                self.wfile.write(self.data)
                return

            def log_message(self, *args, **kwargs):
                return

        server_address = ('127.0.0.1', 0)
        httpd = http_server.HTTPServer(server_address,
                                       BinaryHTTPRequestHandler)
        port = httpd.server_address[1]

        def serve_requests(server):
            server.serve_forever()

        thread = threading.Thread(target=serve_requests, args=(httpd,))
        thread.daemon = True
        thread.start()
        uri = 'http://localhost:%d/' % port
        return (httpd, thread, uri)

    def test_decompress_gzip_with_glance_direct(self):
        """Test GZIP decompression during glance-direct import."""
        self.start_server()

        original_content = b'X' * 10000
        compressed_file = self._create_gzip_file(original_content)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='gzip-image',
            disk_format='raw',
            container_format='bare')

        with open(compressed_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        self.api_methods.verify_image_size(image_id, expected_size=10000)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        expected_checksum = self._calculate_checksum(original_content)
        expected_sha512 = self._calculate_sha512(original_content)
        self.assertEqual(expected_checksum, image['checksum'])
        self.assertEqual(expected_sha512, image['os_hash_value'])

        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        downloaded_data = response.body
        self.assertEqual(10000, len(downloaded_data))
        self.assertEqual(original_content, downloaded_data)

    def test_decompress_gzip_with_web_download(self):
        """Test GZIP decompression during web-download import."""
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        original_content = b'Y' * 15000
        compressed_file = self._create_gzip_file(original_content)

        image_id = self.api_methods.create_and_verify_image(
            name='gzip-web-image',
            disk_format='raw',
            container_format='bare')

        with open(compressed_file, 'rb') as f:
            compressed_data = f.read()

        httpd, thread, image_data_uri = self._start_binary_http_server(
            compressed_data)

        data = {'method': {'name': 'web-download', 'uri': image_data_uri}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        self.api_methods.verify_image_size(image_id, expected_size=15000)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        expected_checksum = self._calculate_checksum(original_content)
        self.assertEqual(expected_checksum, image['checksum'])

        httpd.shutdown()
        httpd.server_close()

    def test_decompress_zip_with_glance_direct(self):
        """Test ZIP decompression during glance-direct import."""
        self.start_server()

        original_content = b'Z' * 20000
        zip_file = self._create_zip_file(original_content)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='zip-image',
            disk_format='raw',
            container_format='bare')

        with open(zip_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        self.api_methods.verify_image_size(image_id, expected_size=20000)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        expected_checksum = self._calculate_checksum(original_content)
        self.assertEqual(expected_checksum, image['checksum'])

        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        downloaded_data = response.body
        self.assertEqual(20000, len(downloaded_data))
        self.assertEqual(original_content, downloaded_data)

    def test_decompress_zip_with_web_download(self):
        """Test ZIP decompression during web-download import."""
        self.config(allowed_ports=[], group='import_filtering_opts')
        self.start_server()

        original_content = b'W' * 12000
        zip_file = self._create_zip_file(original_content)

        image_id = self.api_methods.create_and_verify_image(
            name='zip-web-image',
            disk_format='raw',
            container_format='bare')

        with open(zip_file, 'rb') as f:
            compressed_data = f.read()

        httpd, thread, image_data_uri = self._start_binary_http_server(
            compressed_data)

        data = {'method': {'name': 'web-download', 'uri': image_data_uri}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        self.api_methods.verify_image_size(image_id, expected_size=12000)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        expected_checksum = self._calculate_checksum(original_content)
        self.assertEqual(expected_checksum, image['checksum'])

        httpd.shutdown()
        httpd.server_close()

    def test_decompress_zip_multiple_files_error(self):
        """Test that ZIP with multiple files fails during import."""
        self.start_server()

        # Create a ZIP file with multiple files
        files_dict = {
            'file1.raw': b'content1',
            'file2.raw': b'content2'
        }
        zip_file = self._create_zip_file_multiple(files_dict)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='zip-multi-image',
            disk_format='raw',
            container_format='bare')

        # Stage the compressed file
        with open(zip_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        self._wait_for_task_failure(image_id)

        # Verify image state after task failure
        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        self.assertEqual('importing', image['status'])
        self.assertIsNone(image.get('checksum'))

    def test_skip_decompression_container_format_compressed(self):
        """Test decompression is skipped when container_format='compressed'."""
        self.start_server()

        original_content = b'C' * 8000
        compressed_file = self._create_gzip_file(original_content)
        compressed_size = os.path.getsize(compressed_file)

        image_id = self.api_methods.create_and_verify_image(
            name='compressed-format-image',
            disk_format='raw',
            container_format='compressed')

        with open(compressed_file, 'rb') as f:
            compressed_data = f.read()
            self.api_methods.stage_image_data(image_id, data=compressed_data)

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        self.assertEqual(compressed_size, image['size'])

        expected_checksum = self._calculate_checksum(compressed_data)
        self.assertEqual(expected_checksum, image['checksum'])

        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        downloaded_data = response.body
        self.assertEqual(compressed_size, len(downloaded_data))
        self.assertEqual(compressed_data, downloaded_data)

    def test_no_decompression_uncompressed_file(self):
        """Test that uncompressed files pass through unchanged."""
        self.start_server()

        original_content = b'U' * 5000
        uncompressed_file = self._create_uncompressed_file(original_content)

        image_id = self.api_methods.create_and_verify_image(
            name='uncompressed-image',
            disk_format='raw',
            container_format='bare')

        with open(uncompressed_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        func_utils.wait_for_status(
            self, request_path='/v2/images/%s' % image_id,
            request_headers=self._headers(),
            status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)

        self.api_methods.verify_image_size(image_id, expected_size=5000)

        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        expected_checksum = self._calculate_checksum(original_content)
        self.assertEqual(expected_checksum, image['checksum'])

        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        downloaded_data = response.body
        self.assertEqual(5000, len(downloaded_data))
        self.assertEqual(original_content, downloaded_data)

    def _create_lha_file(self, content, filename='test.lha',
                         entry_name='test_image.raw'):
        """Create an LHA file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with open(filepath, 'wb') as f:
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)
        return filepath

    @testtools.skipIf(
        image_decompression.NO_LHA,
        "lhafile library not available")
    def test_decompress_lha_with_glance_direct(self):
        """Test LHA decompression during glance-direct import."""
        self.start_server()

        lha_file = self._create_lha_file(b'H' * 9000)

        # Create an image
        image_id = self.api_methods.create_and_verify_image(
            name='lha-image',
            disk_format='raw',
            container_format='bare')

        with open(lha_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        # If LHA file is not valid, import will fail.
        try:
            func_utils.wait_for_status(
                self, request_path='/v2/images/%s' % image_id,
                request_headers=self._headers(),
                status='active', max_sec=40, delay_sec=0.2, start_delay_sec=1)
            # If import succeeded, verify final status is active
            path = '/v2/images/%s' % image_id
            response = self.api_get(path, headers=self._headers())
            self.assertEqual(http.OK, response.status_code)
            image = jsonutils.loads(response.text)
            self.assertEqual('active', image['status'])
        except Exception:
            # If import failed, verify it's in importing status with no
            # checksum
            path = '/v2/images/%s' % image_id
            response = self.api_get(path, headers=self._headers())
            self.assertEqual(http.OK, response.status_code)
            image = jsonutils.loads(response.text)
            self.assertEqual('importing', image['status'])
            self.assertIsNone(image.get('checksum'))

    def test_decompress_lha_no_library_error(self):
        """Test LHA import fails gracefully when library is unavailable."""
        if not image_decompression.NO_LHA:
            self.skipTest("lhafile library is available, cannot test NO_LHA")

        self.start_server()

        lha_file = self._create_lha_file(b'LHA content')

        image_id = self.api_methods.create_and_verify_image(
            name='lha-no-lib-image',
            disk_format='raw',
            container_format='bare')

        with open(lha_file, 'rb') as f:
            self.api_methods.stage_image_data(image_id, data=f.read())

        data = {'method': {'name': 'glance-direct'}}
        self.api_methods.import_image(image_id, data=data)

        self._wait_for_task_failure(image_id)

        # Verify image state after task failure
        path = '/v2/images/%s' % image_id
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        image = jsonutils.loads(response.text)
        self.assertEqual('importing', image['status'])
        self.assertIsNone(image.get('checksum'))



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import gzip
import os
from unittest import mock
import zipfile

import glance.async_.flows.plugins.image_decompression as image_decompression
from glance import gateway
import glance.tests.utils as test_utils

UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestDecompressImageTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestDecompressImageTask, self).setUp()

        self.context = mock.MagicMock()
        self.img_repo = mock.MagicMock()
        self.image_id = UUID1
        self.task_id = 'fake-task-id'
        self.task_type = 'api_image_import'

        self.gateway = gateway.Gateway()
        self.img_factory = self.gateway.get_image_factory(self.context)
        self.image = self.img_factory.new_image(image_id=self.image_id,
                                                disk_format='raw',
                                                container_format='bare')
        self.image.size = 1000  # Initial compressed size
        self.img_repo.get.return_value = self.image

    def _create_gzip_file(self, content, filename='test.gz'):
        """Create a gzipped file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with gzip.open(filepath, 'wb') as f:
            f.write(content)
        return filepath

    def test_decompress_gzip_updates_image_size(self):
        """Test that image size is updated after successful GZIP
        decompression.
        """
        # Create a gzipped file
        original_content = b'x' * 10000  # 10KB uncompressed
        compressed_file = self._create_gzip_file(original_content, 'test.gz')

        # Get compressed size
        compressed_size = os.path.getsize(compressed_file)

        # Create decompression task
        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        # Execute decompression
        file_path = 'file://%s' % compressed_file
        result = decompress_task.execute(file_path)

        # Verify the image size was updated to decompressed size
        self.assertEqual(10000, self.image.size)
        # Verify image_repo.save was called
        self.img_repo.save.assert_called_once_with(self.image)
        # Verify the result is the file path
        self.assertEqual(file_path, result)
        # Verify the file was replaced by checking the size changed.
        self.assertNotEqual(compressed_size, os.path.getsize(compressed_file))
        self.assertEqual(10000, os.path.getsize(compressed_file))

    def test_decompress_no_compression_detected(self):
        """Test that image size is not updated when no compression
        is detected.
        """
        # Create a non-compressed file
        uncompressed_file = os.path.join(self.test_dir, 'test.raw')
        with open(uncompressed_file, 'wb') as f:
            f.write(b'x' * 5000)

        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        file_path = 'file://%s' % uncompressed_file
        result = decompress_task.execute(file_path)

        # Verify image size was NOT updated (no compression detected)
        self.assertEqual(1000, self.image.size)
        # Verify image_repo.save was NOT called
        self.img_repo.save.assert_not_called()
        # Verify the result is the file path
        self.assertEqual(file_path, result)

    def test_decompress_skips_when_container_format_compressed(self):
        """Test that decompression is skipped when container_format
        is 'compressed'.
        """
        # Create a gzipped file
        original_content = b'x' * 10000
        compressed_file = self._create_gzip_file(original_content, 'test.gz')

        # Set container_format to 'compressed'
        self.image.container_format = 'compressed'

        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        file_path = 'file://%s' % compressed_file
        result = decompress_task.execute(file_path)

        # Verify image size was NOT updated
        self.assertEqual(1000, self.image.size)
        # Verify image_repo.save was NOT called
        self.img_repo.save.assert_not_called()
        # Verify the result is the file path
        self.assertEqual(file_path, result)

    @mock.patch('glance.async_.flows.plugins.image_decompression.LOG')
    def test_decompress_logs_size_update(self, mock_log):
        """Test that size update is logged."""
        # Create a gzipped file
        original_content = b'x' * 5000
        compressed_file = self._create_gzip_file(original_content, 'test.gz')

        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        file_path = 'file://%s' % compressed_file
        result = decompress_task.execute(file_path)

        # Verify the result is the file path
        self.assertEqual(file_path, result)
        # Verify logging was called
        mock_log.info.assert_called()
        # Check that the log message contains the expected information
        log_calls = [str(call) for call in mock_log.info.call_args_list]
        size_update_logged = any(
            'Updated image' in str(call) and
            'size to' in str(call) and
            'after decompression' in str(call)
            for call in log_calls)
        self.assertTrue(size_update_logged,
                        "Size update log message not found")

    def test_decompress_handles_missing_file(self):
        """Test that decompression raises FileNotFoundError for missing
        file.
        """
        missing_file = os.path.join(self.test_dir, 'nonexistent.gz')

        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        file_path = 'file://%s' % missing_file
        # The code doesn't check for file existence, so it will raise
        # FileNotFoundError when trying to open the file
        self.assertRaises(FileNotFoundError,
                          decompress_task.execute,
                          file_path)

        # Verify image size was NOT updated
        self.assertEqual(1000, self.image.size)
        # Verify image_repo.save was NOT called
        self.img_repo.save.assert_not_called()

    @mock.patch('glance.async_.flows.plugins.image_decompression.lhafile',
                create=True)
    def test_decompress_lha_updates_image_size(self, mock_lhafile):
        """Test that image size is updated after successful LHA
        decompression.
        """
        # Create a fake LHA file (just a file with LHA magic bytes)
        lha_file = os.path.join(self.test_dir, 'test.lha')
        # Write LHA magic bytes at offset 2: 2d 6c 68
        with open(lha_file, 'wb') as f:
            # Magic bytes + padding
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)

        # Mock lhafile.LhaFile
        mock_lha = mock.MagicMock()
        mock_lha.namelist.return_value = ['test_image.raw']
        mock_lha.read.return_value = b'x' * 15000  # 15KB decompressed content
        mock_lha.fp = mock.MagicMock()  # Mock file pointer
        mock_lhafile.LhaFile.return_value = mock_lha
        mock_lhafile.NO_LHA = False

        # Set NO_LHA to False to enable LHA support
        original_no_lha = image_decompression.NO_LHA
        image_decompression.NO_LHA = False

        try:
            decompress_task = image_decompression._DecompressImage(
                self.context, self.task_id, self.task_type,
                self.img_repo, self.image_id)

            file_path = 'file://%s' % lha_file
            result = decompress_task.execute(file_path)

            # Verify LhaFile was instantiated (not with context manager)
            mock_lhafile.LhaFile.assert_called_once_with(
                lha_file, 'r')
            # Verify namelist was called
            mock_lha.namelist.assert_called_once()
            # Verify read was called (not extract)
            mock_lha.read.assert_called_once_with('test_image.raw')
            # Verify file pointer was closed
            mock_lha.fp.close.assert_called_once()
            # Verify the image size was updated to decompressed size
            self.assertEqual(15000, self.image.size)
            # Verify image_repo.save was called
            self.img_repo.save.assert_called_once_with(self.image)
            # Verify the result is the file path
            self.assertEqual(file_path, result)
        finally:
            image_decompression.NO_LHA = original_no_lha

    @mock.patch('glance.async_.flows.plugins.image_decompression.lhafile',
                create=True)
    def test_decompress_lha_handles_multiple_files_error(self, mock_lhafile):
        """Test that LHA decompression raises error for archives with
        multiple files.
        """
        lha_file = os.path.join(self.test_dir, 'test.lha')
        with open(lha_file, 'wb') as f:
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)

        mock_lha = mock.MagicMock()
        # Multiple files
        mock_lha.namelist.return_value = ['file1.raw', 'file2.raw']
        mock_lhafile.LhaFile.return_value = mock_lha
        mock_lhafile.NO_LHA = False

        original_no_lha = image_decompression.NO_LHA
        image_decompression.NO_LHA = False

        try:
            decompress_task = image_decompression._DecompressImage(
                self.context, self.task_id, self.task_type,
                self.img_repo, self.image_id)

            file_path = 'file://%s' % lha_file
            # Should raise Exception for multiple files
            try:
                decompress_task.execute(file_path)
                self.fail("Expected Exception for multiple files")
            except Exception as e:
                self.assertIn('more than one file', str(e))

            # Verify file pointer was still closed in finally block
            mock_lha.fp.close.assert_called_once()
            # Verify image size was NOT updated
            self.assertEqual(1000, self.image.size)
            # Verify image_repo.save was NOT called
            self.img_repo.save.assert_not_called()
        finally:
            image_decompression.NO_LHA = original_no_lha

    @mock.patch('glance.async_.flows.plugins.image_decompression.lhafile',
                create=True)
    def test_decompress_lha_handles_no_lha_library(self, mock_lhafile):
        """Test that LHA decompression raises error when lhafile library
        is not available.
        """
        lha_file = os.path.join(self.test_dir, 'test.lha')
        with open(lha_file, 'wb') as f:
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)

        original_no_lha = image_decompression.NO_LHA
        image_decompression.NO_LHA = True

        try:
            decompress_task = image_decompression._DecompressImage(
                self.context, self.task_id, self.task_type,
                self.img_repo, self.image_id)

            file_path = 'file://%s' % lha_file
            # Should raise Exception when NO_LHA is True
            try:
                decompress_task.execute(file_path)
                self.fail("Expected Exception when NO_LHA is True")
            except Exception as e:
                self.assertIn('No lhafile available', str(e))

            # Verify LhaFile was NOT called
            mock_lhafile.LhaFile.assert_not_called()
            # Verify image size was NOT updated
            self.assertEqual(1000, self.image.size)
            # Verify image_repo.save was NOT called
            self.img_repo.save.assert_not_called()
        finally:
            image_decompression.NO_LHA = original_no_lha

    @mock.patch('glance.async_.flows.plugins.image_decompression.lhafile',
                create=True)
    def test_decompress_lha_closes_file_pointer_on_error(self, mock_lhafile):
        """Test that file pointer is closed even when an error occurs."""
        lha_file = os.path.join(self.test_dir, 'test.lha')
        with open(lha_file, 'wb') as f:
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)

        mock_lha = mock.MagicMock()
        mock_lha.namelist.return_value = ['test_image.raw']
        mock_lha.read.side_effect = IOError("Read error")
        mock_lha.fp = mock.MagicMock()
        mock_lhafile.LhaFile.return_value = mock_lha
        mock_lhafile.NO_LHA = False

        original_no_lha = image_decompression.NO_LHA
        image_decompression.NO_LHA = False

        try:
            decompress_task = image_decompression._DecompressImage(
                self.context, self.task_id, self.task_type,
                self.img_repo, self.image_id)

            file_path = 'file://%s' % lha_file
            # Should raise IOError
            self.assertRaises(IOError, decompress_task.execute, file_path)

            # Verify file pointer was closed in finally block
            mock_lha.fp.close.assert_called_once()
        finally:
            image_decompression.NO_LHA = original_no_lha

    @mock.patch('glance.async_.flows.plugins.image_decompression.lhafile',
                create=True)
    def test_decompress_lha_handles_no_file_pointer(self, mock_lhafile):
        """Test that LHA decompression works when file pointer
        doesn't exist.
        """
        lha_file = os.path.join(self.test_dir, 'test.lha')
        with open(lha_file, 'wb') as f:
            f.write(b'\x00\x00\x2d\x6c\x68' + b'X' * 100)

        mock_lha = mock.MagicMock()
        mock_lha.namelist.return_value = ['test_image.raw']
        mock_lha.read.return_value = b'x' * 8000
        mock_lha.fp = None  # No file pointer
        mock_lhafile.LhaFile.return_value = mock_lha
        mock_lhafile.NO_LHA = False

        original_no_lha = image_decompression.NO_LHA
        image_decompression.NO_LHA = False

        try:
            decompress_task = image_decompression._DecompressImage(
                self.context, self.task_id, self.task_type,
                self.img_repo, self.image_id)

            file_path = 'file://%s' % lha_file
            result = decompress_task.execute(file_path)

            # Should work fine even without file pointer
            self.assertEqual(8000, self.image.size)
            self.img_repo.save.assert_called_once_with(self.image)
            self.assertEqual(file_path, result)
        finally:
            image_decompression.NO_LHA = original_no_lha

    def test_get_flow_returns_flow_with_decompress_task(self):
        """Test that get_flow returns a flow with the decompress task."""
        kwargs = {
            'context': self.context,
            'task_id': self.task_id,
            'task_type': self.task_type,
            'image_repo': self.img_repo,
            'image_id': self.image_id
        }

        flow = image_decompression.get_flow(**kwargs)

        # Verify flow is created
        self.assertIsNotNone(flow)
        # Verify flow is a taskflow Flow object
        from taskflow.patterns import linear_flow as lf
        self.assertIsInstance(flow, lf.Flow)
        # Verify flow has tasks (check via iterating)
        tasks = list(flow)
        self.assertEqual(1, len(tasks))
        task = tasks[0]
        self.assertIsInstance(task, image_decompression._DecompressImage)
        self.assertEqual(self.image_id, task.image_id)
        self.assertEqual(self.task_id, task.task_id)
        self.assertEqual(self.task_type, task.task_type)

    def _create_zip_file(self, content, filename='test.zip',
                         entry_name='test_image.raw'):
        """Create a ZIP file with the given content."""
        filepath = os.path.join(self.test_dir, filename)
        with zipfile.ZipFile(filepath, 'w') as zf:
            zf.writestr(entry_name, content)
        return filepath

    def test_decompress_zip_updates_image_size(self):
        """Test that image size is updated after successful ZIP
        decompression.
        """
        # Create a ZIP file with single file
        original_content = b'x' * 12000  # 12KB uncompressed
        zip_file = self._create_zip_file(original_content, 'test.zip',
                                         'test_image.raw')

        # Get compressed size
        compressed_size = os.path.getsize(zip_file)

        # Create decompression task
        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        # Execute decompression
        file_path = 'file://%s' % zip_file
        result = decompress_task.execute(file_path)

        # Verify the image size was updated to decompressed size
        self.assertEqual(12000, self.image.size)
        # Verify image_repo.save was called
        self.img_repo.save.assert_called_once_with(self.image)
        # Verify the result is the file path
        self.assertEqual(file_path, result)
        # Verify the file was decompressed (original file replaced)
        self.assertTrue(os.path.exists(zip_file))
        # Verify compressed size is different from decompressed size
        self.assertNotEqual(compressed_size, os.path.getsize(zip_file))
        self.assertEqual(12000, os.path.getsize(zip_file))

    def test_decompress_zip_handles_multiple_files_error(self):
        """Test that ZIP decompression raises error for archives with
        multiple files.
        """
        zip_file = os.path.join(self.test_dir, 'test.zip')
        with zipfile.ZipFile(zip_file, 'w') as zf:
            zf.writestr('file1.raw', b'content1')
            zf.writestr('file2.raw', b'content2')

        decompress_task = image_decompression._DecompressImage(
            self.context, self.task_id, self.task_type,
            self.img_repo, self.image_id)

        file_path = 'file://%s' % zip_file
        # Should raise Exception for multiple files
        try:
            decompress_task.execute(file_path)
            self.fail("Expected Exception for multiple files")
        except Exception as e:
            self.assertIn('more than one file', str(e))

        # Verify image size was NOT updated
        self.assertEqual(1000, self.image.size)
        # Verify image_repo.save was NOT called
        self.img_repo.save.assert_not_called()
